__all__ = ["FileFits", "test_fits_magic"]

from .datafile import DataFile
from astropy.io import fits
from a99 import overwrite_fits


class FileFits(DataFile):
    """
    Generic representation of a FITS file

    **Note** normally, the DataFile classes load operation reads all contents
    and closes the file, however this class keeps the file **open** as an
    astropy.io.fits.HDUList object in self.hdulist, because apparently many
    HDUList methods need the file open to work, even after calling
    HDUList.readall()
    """

    # Take this out of explorer.py, too generic
    flag_collect = False
    flag_txt = False
    attrs = ["hdulist"]

    def __init__(self):
        DataFile.__init__(self)
        self.hdulist = None

    def _do_load(self, filename):
        self.hdulist = fits.open(filename)
        self.filename = filename

    def _do_save_as(self, filename):
        """Saves HDU list to FITS file."""
        overwrite_fits(self.hdulist, filename)


    def _test_magic(self, filename):
        test_fits_magic(filename)


def test_fits_magic(filename):
    """For FITS files, we assume that they must start with "SIMPLE".

    This may not be specified in the standard, but I have seen this in all the files I opened"""

    with open(filename, "rb") as file:
        chunk = file.read(4096 * 4)

        if not b"SIMPLE" in chunk:
            raise RuntimeError("File '{}' does not appear to be a FITS file".format(filename))



from astropy.io import fits
from astropy import units as u
from scipy import interp
import numpy as np
import a99
import os

__all__ = ["Spectrum", "fnu", "flambda"]


# AstroPy units defined for convenience
fnu = u.erg/u.cm**2/u.s/u.Hz
flambda = u.erg/u.cm**2/u.s/u.angstrom


@a99.froze_it
class Spectrum(object):
    """
    Spectrum with several *in-place* operations, conversion utilities etc.

    Routines that have the term *in place*: it means that the spectrum
    itself is changed and nothing is returned
    """

    @property
    def xunit(self):
        """Unit of the wavelength axis (some unit from astropy.units)"""
        return self.more_headers["X-UNIT"]

    @xunit.setter
    def xunit(self, value):
        if not isinstance(value, (u.UnitBase,)):  #(u.Unit, u.CompositeUnit, u.IrreducibleUnit)):
            raise TypeError("Spectrum x-unit must be an astropy unit, not '{}'".format(str(value)))
        self.more_headers["X-UNIT"] = value

    @property
    def yunit(self):
        """Unit of the wavelength axis (some unit from astropy.units)"""
        return self.more_headers["Y-UNIT"]

    @yunit.setter
    def yunit(self, value):
        if not isinstance(value, (u.UnitBase,)):  #, (u.Unit, u.CompositeUnit)):
            raise TypeError("Spectrum y-unit must be an astropy unit, not '{}'".format(str(value)))
        self.more_headers["Y-UNIT"] = value

    @property
    def title(self):
        """"Title" of the spectrum. Returns self._title or if None, other options"""
        ret = self._title
        if ret is None and self.filename is not None:
            ret = os.path.split(self.filename)[1]
        if ret is None:
            ret = self.more_headers.get("ORIGIN")
        return ret

    @title.setter
    def title(self, value):
        self._title = value

    @property
    def wavelength(self):
        """Wavelength vector. Equivalent to `wavelenght`"""
        return self.x

    @wavelength.setter
    def wavelength(self, x):
        self.x = x

    @property
    def l0(self):
        """Lower bound of wavelength vector"""
        return self.x[0]

    @property
    def lf(self):
        """Upper bound of wavelength vector"""
        return self.x[-1]

    @property
    def llzero(self):
        """Lower bound of wavelength vector. Synonym with l0"""
        return self.x[0]

    @property
    def llfin(self):
        """Upper bound of wavelength vector. Synonym with lf"""
        return self.x[-1]

    @property
    def flux(self):
        """Flux vector. Equivalent to `y`"""
        return self.y

    @flux.setter
    def flux(self, y):
        self.y = y

    @property
    def delta_lambda(self):
        """Returns delta lambda or NaN if found not to be constant"""
        ret = self.x[1]-self.x[0]
        if abs(self.x[-1]-self.x[-2]-ret) > 1e-5:
            return float("nan")
        return ret

    @property
    def pixel_x(self):
        return self.more_headers.get("PIXEL-X")

    @pixel_x.setter
    def pixel_x(self, value):
        self.more_headers["PIXEL-X"] = value

    @property
    def pixel_y(self):
        return self.more_headers.get("PIXEL-Y")
        
    @pixel_y.setter
    def pixel_y(self, value):
        self.more_headers["PIXEL-Y"] = value

    @property
    def z_start(self):
        return self.more_headers.get("Z-START")

    @z_start.setter
    def z_start(self, value):
        self.more_headers["Z-START"] = value

    # TODO maybe this is obsolete
    @property
    def ylabel(self):
        """Y-label for plotting purposes"""
        return self.more_headers.get("YLABEL")

    @ylabel.setter
    def ylabel(self, value):
        self.more_headers["YLABEL"] = value

    #*****************
    def __init__(self, x=None, y=None):
        self._flag_created_by_block = False  # assertion
        self.more_headers = {}
        self.x = x
        self.y = y
        # Default units
        self.xunit = u.angstrom
        self.yunit = fnu
        self.filename = None
        self.title = None

    def __len__(self):
        """Corresponds to nulbad "ktot"."""
        return len(self.x) if self.x is not None else 0

    def one_liner_str(self):
        if self.x is not None and len(self.x) > 0:
            # s = ", ".join(["%g <= lambda <= %g" % (self.x[0], self.x[-1]),
            #                "delta_lambda = %g" % self.delta_lambda,
            #                "no. points: %d" % len(self.x)])

            info = ["{0:g} \u2264 \u03BB \u2264 {1:g}".format(self.x[0], self.x[-1])]
            try:
                dl = self.delta_lambda
                info.append("\u0394\u03BB = {0:g}".format(dl))
            except:
                pass

            info.append("{0:g} \u2264 flux \u2264 {1:g}".format(np.min(self.y), np.max(self.y)))
            info.append("length: {0:d}".format(len(self.x)))

            s = " | ".join(info)
        else:
            s = "(empty)"
        return s

    def __str__(self):
        return self.one_liner_str()
        # s = ", ".join(["ikeytot = ", str(self.ikeytot), "\n",
        #                "tit = ", str(self.tit), "\n",
        #                "tetaef = ", str(self.tetaef), "\n",
        #                "glog = ", str(self.glog), "\n",
        #                "asalog = ", str(self.asalog), "\n",
        #                "modeles_nhe = ", str(self.modeles_nhe), "\n",
        #                "amg = ", str(self.amg), "\n",
        #                "l0 = ", str(self.l0), "\n",
        #                "lf = ", str(self.lf), "\n",
        #                "pas = ", str(self.pas), "\n",
        #                "echx = ", str(self.echx), "\n",
        #                "echy = ", str(self.echy), "\n",
        #                "fwhm = ", str(self.fwhm), "\n",
        #                "============\n"
        #                "Size of Spectrum: ", str(len(self)), "\n"])
        # return s

    def clear_more_headers(self):
        """Clears .more_headers dictionary, except for X-UNIT and Y-UNIT"""

        keys = list(self.more_headers.keys())
        for key in keys:
            if key not in ("X-UNIT", "Y-UNIT"):
                del self.more_headers[key]





    # header parameters *not* to be put automatically in self.more_headers
    #                  saved separately  dealt with automatically by the FITS module
    #                  ----------------  ------------------------------------------------
    _IGNORE_HEADERS = ("CRVAL", "CDELT", "NAXIS", "PCOUNT", "BITPIX", "GCOUNT", "XTENSION",
                       "XUNIT", "BUNIT", "SIMPLE", "EXTEND", "X-UNIT", "Y-UNIT", "UNITS2", "UNITS1")
    def from_hdu(self, hdu):
        # x/wavelength and y/flux
        n = hdu.data.shape[0]
        lambda0 = hdu.header["CRVAL1"]
        try:
            delta_lambda = hdu.header["CDELT1"]
        except Exception as E:  # todo figure out the type of exception (KeyError?)
            delta_lambda = hdu.header['CD1_1']
            print("Alternative delta lambda in FITS header: CD1_1")
            print("Please narrow the Exception specification in the code")
            print(("Exception is: " + str(E) + " " + E.__class__.__name__))
            print(delta_lambda)
        self.x = np.linspace(lambda0, lambda0 + delta_lambda * (n - 1), n)
        self.y = hdu.data

        _xunit = hdu.header.get("X-UNIT")
        if _xunit:
            self.xunit = u.Unit(_xunit)
        else:
            _xunit = hdu.header.get("XUNIT")
            if _xunit == "nm":
                self.xunit = u.nm
            elif _xunit == "A":
                self.xunit = u.angstrom
            else:
                #  Fallback to angstrom
                self.xunit = u.angstrom

        _yunit = hdu.header.get("Y-UNIT")
        if _yunit:
            self.yunit = u.Unit(_yunit)
        else:
            _yunit = hdu.header.get("BUNIT")
            if _yunit == "erg/s/cm2/Hz":
                self.yunit = fnu
            elif _yunit == "erg/s/cm2/A":
                self.yunit = flambda
            else:
                # y-units as in file C001231_spintg.fits (a card "UNITS2" with blank value exists,
                # but its comment is "electrons (TOTAL)")
                try:
                    _yunit = hdu.header.comments["UNITS1"]
                except KeyError:
                    try:
                        _yunit = hdu.header.comments["UNITS2"]
                    except KeyError:
                        _yunit = None

                if str(_yunit).startswith("electrons (TOTAL)"):
                    self.yunit = u.electron
                else:
                    # Fallback to dimensionless unit
                    self.yunit = u.Unit("")

        # Additional header fields
        for name in hdu.header:
            if not name.startswith(self._IGNORE_HEADERS):
                # **Note** that it eliminates "#" comments from FITS header
                value = hdu.header[name]
                if isinstance(value, str):
                    value = value.split("#")[0].strip()
                self.more_headers[name] = value

    def to_hdu(self):
        """Converts to FITS HDU. delta_lambda must be a constant"""

        if np.isnan(self.delta_lambda):
            # TODO add "spec" parameter to allow other FITS encoding for spectra
            raise RuntimeError("Cannot encode HDU because delta_lambda is not constant")

        hdu = fits.PrimaryHDU()
        # hdu.header[
        #     "telescop"] = "PFANT synthetic spectrum"  # "suggested" by https://python4astronomers.github.io/astropy/fits.html
        hdu.header["CRVAL1"] = self.x[0]
        hdu.header["CDELT1"] = self.delta_lambda


        _xunit = str(self.xunit)
        _yunit = str(self.yunit)
        hdu.header["X-UNIT"] = _xunit
        hdu.header["Y-UNIT"] = _yunit

        # AOSSS expectations
        _xunit = "(not supported)"
        _yunit = "(not supported)"
        if self.xunit == u.angstrom:
            _xunit = "A"
        elif self.xunit == u.nm:
            _xunit = "nm"
        if self.yunit == fnu:
            _yunit = "erg/s/cm2/Hz"
        elif self.yunit == flambda:
            _yunit = "erg/s/cm2/A"
        hdu.header["XUNIT"] = _xunit
        hdu.header["BUNIT"] = _yunit

        for key, value in list(self.more_headers.items()):
            if key in ("X-UNIT", "Y-UNIT"):
                continue

            try:
                # handles commentary cards
                #
                # http://docs.astropy.org/en/stable/io/fits/usage/headers.html#comment-history-and-blank-keywords
                if key in ("COMMENT", "HISTORY", ""):
                    if isinstance(value, str):
                        # proceeds as usual
                        hdu.header[key] = value
                    else:
                        # value is probably an instance of fits.header._HeaderCommentaryCards
                        for _ in value:
                            hdu.header[key] = _
                else:
                    hdu.header[key] = value
            except:
                a99.get_python_logger().exception("Error adding header['{0!s}'] = '{1!s}'".format(key, value))
                raise
        hdu.data = self.y
        return hdu
        
    def cut(self, l0, l1):
        """Cuts *in place* to wavelength interval [l0, l1]. Cut is done within the array objects thus keeping the same objects"""
        ii_delete = np.where(np.logical_or(self.x < l0, self.x > l1))
        self.x = np.delete(self.x, ii_delete)
        self.y = np.delete(self.y, ii_delete)
        
    def cut_idxs(self, i0, i1):
        """Cuts *in place* to slice i0:i1 (pythonic, interval = [i0, i1["""
        self.x = self.x[i0:i1]
        self.y = self.y[i0:i1]

    def resample(self, new_wavelength, kind='linear'):
        """Resamples *in-place* to new wavelength vector using interpolation

        Args:
            new_wavelength: new wavelength vector
            kind: interpolation kind to be passed to scipy.interpolate.interp1d
        """
        # f = interp1d(self.x, self.y, kind='linear', bounds_error=True, fill_value=0)
        self.y = interp(new_wavelength, self.x, self.y)
        self.wavelength = new_wavelength


    def convert_x(self, new_unit):
        """Converts x-axis to new unit"""

    # def flambda_to_fnu(self):
    #     """
    #     Flux-nu to flux-lambda conversion **in-place**
    #
    #     Formula:
    #         f_nu = f_lambda*(lambda/nu) = f_lambda*lambda**2/c
    #
    #         where
    #             lambda is the wavelength in cm,
    #             c is the speed of light in cm/s
    #             f_lambda has irrelevant unit for this purpose
    #
    #     **Note** By convention in this library, Spectrum wavelength is always in angstrom
    #     """
    #
    #     x_cm = self.x*1e-8
    #     self.y *= 1./(1e-8 * C) * x_cm ** 2
    #
    # def fnu_to_flambda(self):
    #     """
    #
    #     **in-place** flux-lambda to flux-nu conversion
    #
    #     Converts flux from erg/s/cm**2/Hz
    #                     to erg/s/cm**2/angstrom
    #     Formula:
    #         f_lambda = f_nu*(nu/lambda) = f_nu*c/lambda**2
    #
    #         (terms description are the same as in flambda_to_fnu())
    #     """
    #
    #     x_cm = self.x * 1e-8  # angstrom -to cm
    #     #    y * C/x_cm ** 2  would be in erg/s/cm**2/cm
    #     # 1e-8 * C/x_cm ** 2        is in erg/s/cm**2/angstrom
    #     self.y *= 1e-8 * C/x_cm ** 2



import sys
import os
import shutil

__all__ = ["get_default_data_path", "copy_default_data_file"]


def get_default_data_path(*args, module=None, class_=None, flag_raise=True):
    """
    Returns path to default data directory

    Arguments 'module' and 'class' give the chance to return path relative to package other than
    f311.filetypes

    Args:
        module: Python module object. It is expected that this module has a sub-subdirectory
                named 'data/default'
        class_: Python class object to extract path information from. If this argument is used,
                it will be expected that the class "root" package will have a sub-subdirectory
                named 'data/default'. Argument 'class_' **has precedence over argument 'module'**
        flag_raise: raises error if file is not found. This can be turned off for whichever purpose
    """
    if module is None:
        module = __get_filetypes_module()

    if class_ is not None:
        pkgname =  class_.__module__
        mseq = pkgname.split(".")
        if len(mseq) < 2 or mseq[1] != "filetypes":
            raise ValueError("Invalid module name for class '{}': '{}' "
                             "(must be '(...).filetypes[.(...)]')".format(class_.__name__, pkgname))
        # gets "root" module object
        # For example, if pkgname is "pyfant.filetypes.filemain", module below will be
        # the "pyfant" module object
        module = sys.modules[mseq[0]]
    module_path = os.path.split(module.__file__)[0]
    p = os.path.abspath(os.path.join(module_path, "data", "default", *args))

    if flag_raise:
        if not os.path.isfile(p):
            raise RuntimeError("Path not found '{}'".format(p))

    return p


def copy_default_data_file(filename, module=None):
    """Copies file from default data directory to local directory."""
    if module is None:
        module = __get_filetypes_module()
    fullpath = get_default_data_path(filename, module=module)
    shutil.copy(fullpath, ".")


def __get_filetypes_module():
    from f311 import filetypes as ft
    return ft




import datetime
from .datafile import DataFile
import a99
import re

__all__ = ["FilePy", "ConfigDict", "FilePyConfig"]


@a99.froze_it
class ConfigDict(dict):
    """Data class to store config options as dictionary"""
    def __missing__(self, key):
        return None

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, dict.__repr__(self))


class FilePy(DataFile):
    """
    Configuration file saved as a .py Python source script

    This class is not intendend to be instantialized. It is an ancestor class for other classes.
    """

    def _test_magic(self, filename):
        with open(filename, "r") as file:
            line = file.readline()
            if not re.match("\s*#\s*-\*-\s*FilePy:\s*{}\s*-\*-".format(self.classname), line):
                raise RuntimeError("File '{}' does not appear to be a '{}' (expected first line of code:"
                                   " \"{}\")".format(filename, self.classname, self._get_magic()))

    def _get_header(self):
        """
        Returns string to be at top of file"""

        return "{}\n#\n# @ Now @ {}\n#\n".format(self._get_magic(), datetime.datetime.now())


    def _get_magic(self):
        """
        Returns string to be written the first line of .py file

        **Note** Newline character **not** included
        """

        return "# -*- FilePy: {} -*-".format(self.classname)

    def _copy_attr(self, module, varname, cls, attrname=None):
        """
        Copies attribute from module object to self. Raises if object not of expected class

        Args:
            module: module object
            varname: variable name
            cls: expected class of variable
            attrname: attribute name of self. Falls back to varname
        """

        if not hasattr(module, varname):
            raise RuntimeError("Variable '{}' not found".format(varname))

        obj = getattr(module, varname)

        if not isinstance(obj, cls):
            raise RuntimeError(
                "Expecting fobj to be a {}, not a '{}'".format(cls.__name__, obj.__class__.__name__))

        if attrname is None:
            attrname = varname

        setattr(self, attrname, obj)



class FilePyConfig(FilePy):
    """Base class for config files. Inherit and set class variable 'modulevarname' besides usual"""

    attrs = ["obj"]

    # Name of variable in module
    modulevarname = "obj"

    def __init__(self):
        FilePy.__init__(self)
        self.obj = ConfigDict()

    def _do_load(self, filename):
        module = a99.import_module(filename)
        self._copy_attr(module, self.modulevarname, ConfigDict, "obj")

    def _do_save_as(self, filename):
        with open(filename, "w") as h:
            h.write("{}\n"
                    "from f311.filetypes import ConfigDict\n"
                    "\n"
                    "{} = {}\n".format(self._get_header(), self.modulevarname, a99.make_code_readable(repr(self.obj))))

    def init_default(self):
        # Already created OK
        pass



"""
API in functional form for plotting spectra

Routines also used by plot-spectra.py
"""

import matplotlib.pyplot as plt
import math
import matplotlib.backends.backend_pdf
from matplotlib import rc
import logging
import numpy as np
import a99
import f311.filetypes as ft

__all__ = ["plot_spectra_stacked", "plot_spectra_overlapped", "plot_spectra_pieces_pdf",
 "plot_spectra_pages_pdf", "draw_spectra_stacked", "PlotSpectrumSetup", "draw_spectra_overlapped",
 "calc_max_min"]


_T = 0.02  # percentual amount of extra space on left, right, top, bottom of graphics
_FAV_COLOR = 'k'  # "favourite color" for single-spectrum plots


# TODO draw_spectra_overlapped (maybe when I need it)


def _set_plot(callable_, fmt, spectrum):
    """Sets something in plot calling callable_(fmt.format(spectrum) if fmt is not empty"""
    if fmt:
        callable_(fmt.format(spectrum))

class PlotSpectrumSetup(object):
    """

    Args:
        fmt_xlabel: format string for x-label
        fmt_ylabel: format string for y-label
        fmt_title: format string for title
        ymin: (optional) force mininum y-value
        flag_legend: Whether to show legend in plot
    """
    def __init__(self, fmt_xlabel="{.xunit}", fmt_ylabel="{.yunit}", fmt_title="{.title}",
                 ymin=None, flag_legend=True, flag_xlabel=True, flag_ylabel=True):
        self.flag_xlabel = flag_xlabel
        self.flag_ylabel = flag_ylabel
        self.fmt_xlabel = fmt_xlabel
        self.fmt_ylabel = fmt_ylabel
        self.fmt_title = fmt_title
        self.ymin = ymin
        self.flag_legend = flag_legend

    def __repr__(self):
        return "{}('{}', '{}', '{}', {}, {})".format(self.__class__.__name__,self.fmt_xlabel,
          self.fmt_ylabel, self.fmt_title, self.ymin, self.flag_legend)


_default_setup = PlotSpectrumSetup()


def plot_spectra_stacked(ss, title=None, num_rows=None, setup=_default_setup):
    """
    Plots one or more stacked in subplots sharing same x-axis.

    Args:
      ss: list of Spectrum objects
      title=None: window title
      num_rows=None: (optional) number of rows for subplot grid. If not passed,
        num_rows will be the number of plots, and the number of columns will be 1.
        If passed, number of columns is calculated automatically.
      setup: PlotSpectrumSetup object

    """

    draw_spectra_stacked(ss, title, num_rows, setup)
    plt.show()
    # return fig


def plot_spectra_overlapped(ss, title=None, setup=_default_setup):
    """
    Plots one or more spectra in the same plot.

    Args:
      ss: list of Spectrum objects
      title=None: window title
      setup: PlotSpectrumSetup object
    """

    plt.figure()
    draw_spectra_overlapped(ss, title, setup)
    plt.show()


def draw_spectra_overlapped(ss, title=None, setup=_default_setup):
    a99.format_BLB()
    if len(ss) > 0:
        xunit, yunit = None, None
        for i, s in enumerate(ss):
            if xunit is None:
                xunit = s.xunit
            else:
                if xunit != s.xunit:
                    raise RuntimeError("Spectra x-units do not match")

            if yunit is None:
                yunit = s.yunit
            else:
                if yunit != s.yunit:
                    raise RuntimeError("Spectra x-units do not match")

            ax = plt.gca()
            y = s.y
            ax.plot(s.x, y, label=str(s.title))

        # plt.ylabel('({})'.format(yunit))
        # **Note** Takes last spectrum as reference to mount x-label
        if setup.flag_xlabel and setup.fmt_xlabel:
            _set_plot(plt.xlabel, setup.fmt_xlabel, s)
        xmin, xmax, ymin_, ymax, xspan, yspan = calc_max_min(ss)
        ymin = ymin_ if setup.ymin is None else setup.ymin
        plt.xlim([xmin - xspan * _T, xmax + xspan * _T])
        plt.ylim([ymin - yspan * _T, ymax + yspan * _T])
        if setup.flag_legend:
            leg = plt.legend(loc=0)
            a99.format_legend(leg)
        plt.tight_layout()

    if title is not None:
        plt.gcf().canvas.set_window_title(title)


def plot_spectra_pieces_pdf(ss, aint=10, pdf_filename='pieces.pdf', setup=_default_setup):
    """
    Plots spectra, overlapped, in small wavelength intervals into a PDF file,
    one interval per page of the PDF file.

    Args:
      ss: list of Spectrum objects
      aint: wavelength interval for each plot
      pdf_filename: name of output file
      setup: PlotSpectrumSetup object

    **Note** overrides setup.fmt_xlabel; leaves y-labell and title blank
    """

    import f311.explorer as ex

    xmin, xmax, ymin_, ymax, _, yspan = calc_max_min(ss)
    ymin = ymin_ if setup.ymin is None else setup.ymin

    num_pages = int(math.ceil((xmax-xmin)/aint)) # rightmost point may be left out...or not
    # num_spectra = len(ss)

    a99.format_BLB()
    # pdf = matplotlib.backends.backend_pdf.PdfPages(pdf_filename)
    pdf = matplotlib.backends.backend_pdf.PdfPages(pdf_filename)
    logger = a99.get_python_logger()

    for h in range(num_pages):
        fig = plt.figure()
        lambda0 = xmin+h*aint
        lambda1 = lambda0+aint
        logger.info("Printing page {0:d}/{1:d} ([{2:g}, {3:g}])".format(h+1, num_pages, lambda0, lambda1))
        for i, s in enumerate(ss):
            s_cut = ex.cut_spectrum(s, lambda0, lambda1)
            ax = plt.gca()
            ax.plot(s_cut.x, s_cut.y, label=s.title)
        if setup.flag_xlabel and setup.fmt_xlabel:
            plt.xlabel('Wavelength (interval: [{0:g}, {1:g}])'.format(lambda0, lambda1))
        xspan = lambda1-lambda0
        ax.set_xlim([lambda0 - xspan * _T, lambda1 + xspan * _T])
        ax.set_ylim([ymin - yspan * _T, ymax + yspan * _T])
        if setup.flag_legend:
            leg = plt.legend(loc=0)
            a99.format_legend(leg)
        plt.tight_layout()
        pdf.savefig(fig)
        plt.close()

    # for fig in xrange(1, figure().number): ## will open an empty extra figure :(
    #     pdf.savefig( fig )
    pdf.close()
    logger.info("File {0!s} successfully created.".format(pdf_filename))


def plot_spectra_pages_pdf(ss, pdf_filename='pages.pdf', setup=_default_setup):
    """
    Plots spectra into a PDF file, one spectrum per page.

    Splits into several pieces of width

    Args:
      ss: list of Spectrum objects
      pdf_filename: name of output file
    """
    logger = a99.get_python_logger()
    xmin, xmax, ymin_, ymax, xspan, yspan = calc_max_min(ss)
    ymin = ymin_ if setup.ymin is None else setup.ymin
    num_pages = len(ss)
    a99.format_BLB()
    pdf = matplotlib.backends.backend_pdf.PdfPages(pdf_filename)
    for i, s in enumerate(ss):
        title = s.title
        fig = plt.figure()
        plt.plot(s.x, s.y, c=_FAV_COLOR)
        if setup.flag_xlabel and setup.fmt_xlabel:
            _set_plot(plt.xlabel, setup.fmt_xlabel, s)
        if setup.flag_ylabel and setup.fmt_ylabel:
            _set_plot(plt.ylabel, setup.fmt_ylabel, s)
        _set_plot(plt.title, setup.fmt_title, s)
        plt.xlim([xmin-xspan*_T, xmax+xspan*_T])
        plt.ylim([ymin-yspan*_T, ymax+yspan*_T])
        plt.tight_layout()
        plt.subplots_adjust(top=0.94) # workaround for cropped title
        logger.info("Printing page {0:d}/{1:d} ('{2!s}')".format(i+1, num_pages, title))
        pdf.savefig(fig)
        plt.close()
    pdf.close()
    logger.info("File {0!s} successfully created.".format(pdf_filename))


def calc_max_min(ss):
    """"Calculates (x, y) (max, min) for a list of Spectrum objects.

    Returns (xmin, xmax, ymin, ymax, xspan, yspan)
    """
    xmin, xmax, ymin, ymax = 1e38, -1e38, 1e38, -1e38
    for s in ss:
        assert isinstance(s, ft.Spectrum)
        if len(s.x) > 0:
            xmin, xmax = min(min(s.x), xmin), max(max(s.x), xmax)
            ymin, ymax = min(min(s.y), ymin), max(max(s.y), ymax)
    xspan = xmax-xmin
    yspan = ymax - ymin
    return xmin, xmax, ymin, ymax, xspan, yspan


def draw_spectra_stacked(ss, title=None, num_rows=None, setup=_default_setup):
    """Same as plot_spectra_stacked(), but does not call plt.show(); returns figure"""
    n = len(ss)
    assert n > 0, "ss is empty"
    if not num_rows:
        num_rows = n
        num_cols = 1
    else:
        num_cols = int(np.ceil(float(n) / num_rows))
    a99.format_BLB()

    fig, axarr = plt.subplots(num_rows, num_cols, sharex=True, squeeze=False)
    xmin = 1e38
    xmax = -1e38
    i, j = -1, num_cols
    xunit, yunit = None, None
    any_span = False
    for s in ss:
        j += 1
        if j >= num_cols:
            i += 1
            if i >= num_rows:
                break
            j = 0
        assert isinstance(s, ft.Spectrum)
        ax = axarr[i, j]
        y = s.y
        ax.plot(s.x, y)
        ymin_, ymax = ax.get_ylim()
        ymin_now = ymin_ if setup.ymin is None else setup.ymin
        ax.set_ylim([ymin_now, ymin_now + (ymax - ymin_now) * (1 + _T)])  # prevents top of line from being hidden by plot box

        _set_plot(ax.set_ylabel, setup.fmt_ylabel, s)
        _set_plot(ax.set_title, setup.fmt_title, s)

        if len(s.x) > 0:
            xmin, xmax = min(min(s.x), xmin), max(max(s.x), xmax)
            any_span = True

        if xunit is None:
            xunit = s.xunit
        else:
            if xunit != s.xunit:
                raise RuntimeError("Spectra x-units do not match")

        if yunit is None:
            yunit = s.yunit
        else:
            if yunit != s.yunit:
                raise RuntimeError("Spectra x-units do not match")

    if any_span:
        span = xmax - xmin
        ax.set_xlim([xmin - span * _T, xmax + span * _T])
    for j in range(num_cols):
        ax = axarr[num_rows - 1, j]
        if setup.flag_xlabel and setup.fmt_xlabel:
            _set_plot(ax.set_xlabel, setup.fmt_xlabel, s)
    plt.tight_layout()
    if title is not None:
        fig.canvas.set_window_title(title)
    return fig



import a99


__all__ = ["Vis", "VisList"]


class Vis(object):
    """
    Base class for visualizations.

    Those who open multiple figures must call plt.show(), otherwise not.
    """

    # Set the classes accepted by the use() method
    input_classes = ()
    # Fill string with a verb and object, e.g. "plot first record"
    action = ""

    def __init__(self):
        self.title = None
        self.parent_form = None

    def use(self, obj, parent_form=None):
        """Note: if title is None, will be replaced with obj.filename
        """

        if not isinstance(obj, self.input_classes):
            raise RuntimeError('{0!s} cannot handle a {1!s}'.format(self.__class__.__name__, obj.__class__.__name__))

        self.parent_form = parent_form

        if self.title is None:
            self.title = 'file: '+obj.filename
        self._do_use(obj)

    def _do_use(self, obj):
        raise NotImplementedError()


class VisList(Vis):
    """Vis subclass that can handle a list of objects to visualize simultaneously"""

    input_classes = (list)
    # Inform the element classes accepted
    item_input_classes = ()

    def use(self, objs, parent_form=None):
        for obj in objs:
            if not isinstance(obj, self.item_input_classes):
                raise RuntimeError('{0!s} cannot handle a {1!s}'.format(self.__class__.__name__, obj.__class__.__name__))

        self.parent_form = parent_form
        self._do_use(objs)



"""
Miscellanea routines that depend on other modules and sub-packages.

Rule: only 'gui/' modules can import util!!!
"""

import numpy as np
import copy
import a99

__all__ = ["cut_spectrum", ]


def cut_spectrum(sp, l0, lf):
    """
    Cuts spectrum given a wavelength interval, leaving origina intact

    Args:
        sp: Spectrum instance
        l0: initial wavelength
        lf: final wavelength

    Returns:
        Spectrum: cut spectrum
    """

    if l0 >= lf:
        raise ValueError("l0 must be lower than lf")
    idx0 = np.argmin(np.abs(sp.x - l0))
    idx1 = np.argmin(np.abs(sp.x - lf))
    out = copy.deepcopy(sp)
    out.x = out.x[idx0:idx1]
    out.y = out.y[idx0:idx1]
    return out




#!/usr/bin/env python

"""F311 Explorer --  file manager-like application to list, visualize, and edit data files"""

import a99
import sys
import argparse
import logging
import f311


a99.logging_level = logging.INFO
a99.flag_log_file = True

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
    description=__doc__,
    formatter_class=a99.SmartFormatter
    )
    parser.add_argument('dir', type=str, help='directory name', default='.', nargs='?')
    args = parser.parse_args()

    app = a99.get_QApplication([])
    form = f311.XExplorer(None, args.dir)
    form.show()
    a99.place_center(form)
    sys.exit(app.exec_())



#!/usr/bin/env python
"""
Cuts spectrum file to wavelength interval specified

Resulting spectrum Saved in 2-column ASCII format

The interval is [llzero, llfin]
"""

import argparse
import logging
import sys
import a99
import f311

a99.logging_level = logging.INFO

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
     description=__doc__,
     formatter_class=a99.SmartFormatter
     )
    parser.add_argument('llzero', type=float, nargs=1,
     help='lower wavelength boundary (angstrom)')
    parser.add_argument('llfin', type=float, nargs=1,
     help='upper wavelength boundary (angstrom)')
    parser.add_argument('fn_input', type=str, help='input file name', nargs=1)
    parser.add_argument('fn_output', type=str, help='output file name', nargs=1)

    args = parser.parse_args()

    sp = f311.load_spectrum(args.fn_input[0])
    if not sp:
        print("File '{0!s}' not recognized as a spectrum file.".format(args.fn_input[0]))
        sys.exit()

    m = min(sp.x)
    M = max(sp.x)

    print("Original interval: [{0:g}, {1:g}]".format(m, M))
    print("New interval: [{0:g}, {1:g}]".format(args.llzero[0], args.llfin[0]))

    f = f311.FileSpectrumXY()
    f.spectrum = f311.cut_spectrum(sp, args.llzero[0], args.llfin[0])
    f.save_as(args.fn_output[0])
    print("Successfully created file '{0!s}'".format(args.fn_output[0]))



#!/usr/bin/env python
"""
Plots spectra on screen or creates PDF file

It can work in four different modes:

a) grid of sub-plots, one for each spectrum (default mode)
   Example:
   plot-spectra.py flux.norm.nulbad measured.fits

b) single plot with all spectra overlapped ("--ovl" option)
   Example:
   > plot-spectra.py --ovl flux.norm.nulbad measured.fits

c) PDF file with a small wavelength interval per page ("--pieces" option).
   This is useful to flick through a large wavelength range.
   Example:
   > plot-spectra.py --pieces --aint 7 flux.norm.nulbad measured.fits

d) PDF file with one spectrum per page ("--pages" option).
   Example:
   > plot-spectra.py --pages flux.*

Types of files supported:

  - pfant output, e.g., flux.norm;
  - nulbad output, e.g., flux.norm.nulbad;
  - 2-column "lambda-flux" generic text files;
  - FITS files.

"""
import argparse
import a99
import logging
import glob
import f311.explorer as ex
import f311.filetypes as ft
import f311

a99.logging_level = logging.INFO
a99.flag_log_file = True

DEFAULT_FN_OUTPUT = '(plot-spectra-<xxxx>.pdf)'

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
     description=__doc__,
     formatter_class=a99.SmartFormatter
    )

    parser.add_argument('fn', type=str, nargs='+',
     help='name of spectrum file(s) (many types supported) '
          '(wildcards allowed, e.g., "flux.*")')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--ovl', help='Overlapped graphics', action="store_true")
    group.add_argument('--pieces', help='If set, will generate a PDF file with '
     'each page containing one "piece" of the spectra of length'
     'given by the --aint option.', action="store_true")
    group.add_argument('--pages', help='If set, will generate a PDF file with '
     'one spectrum per page', action="store_true")
    # parser.add_argument('--ovl', help='Overlapped graphics', action="store_true")
    # parser.add_argument('--pieces', help='If set, will generate a PDF file with '
    #  'each page containing one "piece" of the spectra of length'
    #  'given by the --aint option.', action="store_true")
    # parser.add_argument('--pages', help='If set, will generate a PDF file with '
    #  'one spectrum per page', action="store_true")
    parser.add_argument('--aint', type=float, nargs='?', default=10,
     help='length of each piece-plot in wavelength units (used only if --pieces)')
    parser.add_argument('--fn_output', nargs='?', default=DEFAULT_FN_OUTPUT, type=str,
                        help='PDF output file name (used only if --pieces)')
    parser.add_argument('--ymin', nargs='?', default='(automatic)', type=str,
     help='Minimum value for y-axis')
    parser.add_argument('-r', '--num_rows', nargs='?', default='(automatic)', type=str,
     help='Number of rows in subplot grid')

    args = parser.parse_args()

    ymin = None if args.ymin == "(automatic)" else float(args.ymin)
    num_rows = None if args.num_rows == "(automatic)" else int(args.num_rows)

    # Compiles list of file names.
    # Each item in args.fn list may have wildcards, and these will be expanded
    # into actual filenames, then duplicates are eliminated
    patterns = args.fn
    _ff = []
    for pattern in patterns:
        _ff.extend(glob.glob(pattern))
    ff = []
    for f in _ff:
        if f not in ff:
            ff.append(f)

    # classes = [FileSpectrumPfant, FileSpectrumNulbad, FileSpectrumXY, FileSpectrumFits]

    ss = []
    flag_ok = False
    for x in ff:
        print("Reading file '{0!s}'...".format(x))
        f = f311.load_with_classes(x, f311.classes_sp())
        if f is None:
            a99.print_error("... type not recognized, sorry")
        else:
            print("... successfully read using reader {0!s}.".format(f.__class__.__name__))
            ss.append(f.spectrum)


    # # Making name for output file
    if args.fn_output == DEFAULT_FN_OUTPUT:
        if len(ss) == 1:
            # If there is only one spectrum, makes output PDF filename using spectrum name
            prefix = "plot-spectra-"+ss[0].filename
        else:
            prefix = "plot-spectra"
        fn_output = a99.new_filename(prefix, "pdf")
    else:
        fn_output = args.fn_output

    if len(ss) == 0:
        a99.print_error("Nothing to plot!")
    else:
        setup = ex.PlotSpectrumSetup(ymin=ymin)
        if len(ss) == 1:
            # No need for legend in plot if there is only one spectrum.
            # Spectrum filename will be part of PDF filename anyway
            setup.flag_legend = False

        if args.pieces:
            ex.plot_spectra_pieces_pdf(ss, aint=args.aint, pdf_filename=fn_output, setup=setup)
        elif args.pages:
            ex.plot_spectra_pages_pdf(ss, pdf_filename=fn_output, setup=setup)
        else:
            if args.ovl:
                ex.plot_spectra_overlapped(ss, "", setup=setup)
            else:
                ex.plot_spectra_stacked(ss, "", setup=setup, num_rows=num_rows)



"""
Collaboration-related routines

1) Class Catalog -- resources to retrieve File* classes by different criteria

2) Script utilities: collect scripts (standalone applications) across collaborator packages
"""


from collections import OrderedDict
import importlib
import a99
import os
import glob
import copy


__all__ = [
    "COLLABORATORS_C", "COLLABORATORS_S",
    "classes_txt", "classes_bin", "classes_sp", "classes_file",
    "get_suitable_vis_classes", "get_suitable_vis_list_classes",
    "get_scripts_path", "get_programs_dict",
    "EXTERNAL_COLLABORATORS"
    ]

# List of Python packages to be considered "external collaborators"
#
# These packages may contribute with:
# - scripts
# - DataFile subclasses
# - Vis subclasses
EXTERNAL_COLLABORATORS = ["pyfant", "aosss", "convmolworks", "ariastro"]


# List of **classes** collaborators packages (**change to add**)
#
COLLABORATORS_C = ["f311"]+EXTERNAL_COLLABORATORS
# List of **script** collaborator packages to look for scripts (**change to add**)
#
__F311 = ["f311"]+["f311."+x for x in a99.get_subpackages_names(os.path.split(__file__)[0])]
COLLABORATORS_S = __F311+EXTERNAL_COLLABORATORS


# **        ****                ******        ****                ******        ****
#   **    **    ******    ******      **    **    ******    ******      **    **    ******    ******
#     ****            ****              ****            ****              ****            ****
#
# Class catalog-related routines


def get_suitable_vis_classes(obj):
    """Retuns a list of Vis classes that can handle obj."""

    ret = []
    for class_ in classes_vis():
        if isinstance(obj, class_.input_classes):
            ret.append(class_)
    return ret


def get_suitable_vis_list_classes(objs):
    """Retuns a list of VisList classes that can handle a list of objects."""

    from f311 import explorer as ex

    ret = []
    for class_ in classes_vis():
        if isinstance(class_, ex.VisList):
            flag_can = True
            for obj in objs:
                if not isinstance(obj, class_.item_input_classes):
                    flag_can = False
                    break
            if flag_can:
                ret.append(class_)
    return ret


def classes_txt():
    """Classes to consider when attempts to load a text file (see load_any_file())"""
    if __flag_first:
        __setup()
    return _classes_txt


def classes_bin():
    """Classes to consider when attempts to load a binary file (see load_any_file())"""
    if __flag_first:
        __setup()
    return _classes_bin


def classes_sp():
    """Classes to consider when attempts to load a spectrum file (see load_spectrum())"""
    if __flag_first:
        __setup()
    return _classes_sp


def classes_file(flag_leaf=False):
    """All known File* classes

    Args:
        flag_leaf: returns only classes that do not have subclasses
                   ("leaf" nodes as in a class tree graph)
    """
    if __flag_first:
        __setup()

    if not flag_leaf:
        return _classes_file

    return [cls for cls in _classes_file if cls not in _classes_file_superclass]


def classes_vis():
    """All known Vis* classes"""
    if __flag_first:
        __setup()
    return _classes_vis


def _collect_classes(m):
    """
    Adds entries to _classes_*

    Args:
        m: module object that must contain the following sub-modules: datatypes, vis
    """
    from f311 import filetypes as ft
    from f311 import explorer as ex

    def _extend(classes, newclasses):
        """Filters out classes already present in list.

        This shouldn't be necessary, but collaborators may accidentally import already loaded
        classes into the datatypes namespace"""
        classes.extend([class_ for class_ in newclasses if class_ not in classes])
        # classes.extend(newclasses)

    file_classes = [class_ for class_ in a99.get_classes_in_module(m, ft.DataFile) if class_.flag_collect]

    # Classes to consider when attempts to load a text file (see load_any_file())
    _extend(_classes_txt, [class_ for class_ in file_classes if class_.flag_txt])

    # Classes to consider when attempts to load a binary file (see load_any_file())
    _extend(_classes_bin, [class_ for class_ in file_classes if not class_.flag_txt])
    # Adds Classes to consider when attempts to load a spectrum file (see load_spectrum())
    _extend(_classes_sp, [class_ for class_ in file_classes if issubclass(class_, ft.FileSpectrum)])
    # All kwown File* classes
    _extend(_classes_file, file_classes)
    # All kwnown Vis* classes

    _extend(_classes_vis, a99.get_classes_in_module(m, ex.Vis))

    global _classes_file_superclass
    _classes_file_superclass = [cls.__bases__[0] for cls in _classes_file]


# # List of classes representing all file formats either read or written
#   ====================================================================
_classes_txt = []
_classes_bin = []
_classes_sp = []
_classes_file = []
_classes_file_superclass = []  # superclasses of items in _classes_file
_classes_vis = []
__flag_first = True
__collaborators = OrderedDict()


def __setup():
    """Will be executed in the first time someone calls classes_*() """
    global __collaborators, __flag_first

    import f311

    __flag_first = False
    for pkgname in f311.COLLABORATORS_C:
        try:
            pkg = importlib.import_module(pkgname)
            a99.get_python_logger().info("Imported collaborator package '{}'".format(pkgname))

            try:
                if hasattr(pkg, "_setup_filetypes"):
                    pkg._setup_filetypes()
                else:
                    _collect_classes(pkg)

                __collaborators[pkgname] = pkg
            except:
                a99.get_python_logger().exception(
                    "Actually, package '{}' gave error".format(pkgname))
                raise
        except:
            a99.get_python_logger().warning("Failed to import package '{}".format(pkgname))
            # raise


# **        ****                ******        ****                ******        ****
#   **    **    ******    ******      **    **    ******    ******      **    **    ******    ******
#     ****            ****              ****            ****              ****            ****
#
# Scripts-related routines

def get_scripts_path(packagename):
    """**Convention** Returns full path to scripts directory"""
    return os.path.join(packagename, "scripts")


# {"packagename0": {"exeinfo": [ExeInfo00, ...], "description": description0}, ...}
# keys in COLLABORATORS_S
__programs_dict = None


def _get_programs_dict():
    """
    Builds and returns programs dictionary

    This will have to import the packages in COLLABORATORS_S in order to get their absolute path.

    Returns:
        dictionary: {"packagename": [ExeInfo0, ...], ...}

    "packagename" examples: "f311.explorer", "numpy"
    """
    global __programs_dict

    if __programs_dict is not None:
        return __programs_dict

    d = __programs_dict = OrderedDict()

    for pkgname in COLLABORATORS_S:
        try:
            package = importlib.import_module(pkgname)
        except ModuleNotFoundError:
            # I think it is better to be silent when a collaborator package is not installed
            continue

        path_ = os.path.join(os.path.split(package.__file__)[0], "scripts")
        bulk = a99.get_exe_info(path_, flag_protected=True)
        d[pkgname] = {"description": a99.get_obj_doc0(package), "exeinfo": bulk}

    return __programs_dict


def get_programs_dict(pkgname_only=None, flag_protected=False):
    """
    Scans COLLABORATORS_S packages for scripts, eventually filtering if arguments passed

    Args:
        pkgname_only: name of single package within COLLABORATORS_S
        flag_protected: include scripts starting with "_"?

    Returns:
        dictionary: {"packagename0": {"exeinfo": [ExeInfo00, ...], "description": description0}, ...}
    """

    ___ret = _get_programs_dict()
    __ret = ___ret if pkgname_only is None else OrderedDict(((pkgname_only, ___ret[pkgname_only]),))
    if flag_protected:
        _ret = __ret
    else:
        _ret = copy.deepcopy(__ret)
        for value in _ret.values():
            value["exeinfo"] = [exeinfo for exeinfo in value["exeinfo"] if not exeinfo.filename.startswith("_")]

    # Removes packages that may have gone out of scripts after filtering
    ret = _ret if pkgname_only is None and flag_protected is None else \
        OrderedDict(((key, value) for key, value in _ret.items() if len(value["exeinfo"]) > 0))

    return ret



from ... import filetypes as ft
from .basic import Vis
import matplotlib.pyplot as plt


__all__ = ["VisPrint", "VisSpectrum"]


class VisPrint(Vis):
    """Prints object to screen."""

    input_classes = (object,)
    action = "Print to console"

    def _do_use(self, obj):
        print(obj)


class VisSpectrum(Vis):
    """Plots single spectrum."""

    input_classes = (ft.FileSpectrum,)
    action = "Plot spectrum"

    def _do_use(self, m):
        from f311 import explorer as ex
        s = m.spectrum
        ex.plot_spectra_stacked([s])




"""
Visualization (including plotting)

This package contains Vis* classes and functional API to draw figures
"""



from .basic import *
from .visprint import *
from .visspectrum import *
from .plotsp import *



"""
Object-oriented framework to handle file types:

  - Base class `FilePy` to support Python sources as config files
  - `FileLiteDB` class wraps a SQLite3 database
  - Automatic file type detection
  - Plugin system to support (read/write/visualize) new file types
"""


from . import gui
from . import vis

from .gui import *
from .vis import *
from .util import *


# # Function to access package-specific config file
def get_config():
    """Returns AAConfigObj object that corresponds to file ~/.MY-PACKAGE-NAME.conf"""
    import a99
    return a99.get_config_obj(".f311.explorer.conf")



"""XText: Window to show text; XHTML: window to show HTML."""
__all__ = ["XText", "XHTML"]

from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import a99


################################################################################
class XText(QMainWindow):
    """
    Args:
      parent=None: nevermind
      text: string
    """

    def __init__(self, parent=None, text="", title=""):
        QMainWindow.__init__(self, parent)

        cw = self.centralWidget = QPlainTextEdit()
        cw.setLineWrapMode(QPlainTextEdit.NoWrap)
        self.setCentralWidget(cw)
        cw.setReadOnly(True)  # allows copy but not editing
        cw.setFont(a99.MONO_FONT)
        cw.setPlainText(text)

        self.setWindowTitle(title)
        self.setGeometry(0, 0, 800, 600)
        a99.place_center(self)

################################################################################
class XHTML(QMainWindow):
    """
    Args:
      parent=None: nevermind
      html: string
    """

    def __init__(self, parent=None, html="", title=""):
        QMainWindow.__init__(self, parent)

        cw = self.centralWidget = QTextEdit()
        self.setCentralWidget(cw)
        cw.setReadOnly(True)  # allows copy but not editing
        cw.setText(html)

        self.setWindowTitle(title)
        self.setGeometry(0, 0, 800, 600)
        a99.place_center(self)




"""Explorer window."""
__all__ = ["XExplorer"]

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import a99
import f311
import os
import glob
import os.path
import time
import matplotlib.pyplot as plt
import numpy as np
# from .. import *


COLOR_LOADED = "#ADFFB4"  # light green
COLOR_LOAD_ERROR = "#FFB5B5"  # light red
COLOR_DIR = "#FFFFB5"  # light yellow
# Maximum size of file to be loaded automatically (in bytes)
MAX_FILESIZE_AUTO_LOAD = 1.2*2**20


@a99.froze_it
class _FileProps(object):
    def __init__(self):
        # DataFile object or None
        self.f = None
        # string or None
        self.filepath = None
        # Already "scanned" (load attempted)
        self.flag_scanned = False
        # Error loading?
        self.flag_error = False
        # Is text?
        self.flag_text = False
        # Is file (name to match os.path.isfile
        self.isfile = False
        # Error message
        self.error_message = ""
        # Row index in table
        self.row_index = None
        # Date changed
        self.mtime = None

    def get_color(self):
        """Returns None if background not to be changed, otherwise a QColor."""
        if not self.isfile:
            return QColor(COLOR_DIR)
        elif not self.flag_scanned:
            return None
        elif self.flag_error:
            return QColor(COLOR_LOAD_ERROR)
        return QColor(COLOR_LOADED)

    def get_summary(self):
        ret = "File not scanned yet"
        if not self.isfile:
            ret = '<span style="color: #B38600">Directory</span>'
        elif self.flag_scanned:
            ll = ["text file" if self.flag_text else "binary file"]
            if self.flag_error:
                emsg = self.error_message or "file type not recognized"
                ll.append("<span style=\"font-weight: bold; color: {0!s}\">{1!s}</span>".format(a99.COLOR_ERROR, emsg))
            else:
                ll.append("<span style=\"color: {0!s}\">File type: <b>{1!s}</b></span>".format("#006000", self.f.__class__.__name__))
            ret = ("; ".join(ll))
        return ret

    def get_info(self):
        ret = ""
        if self.flag_scanned and not self.flag_error:
            ret = str(self.f)
        return ret

class LoadThread(QThread):
    def __init__(self, parent, propss, cls=None):
        # cls: specific class to use
        QThread.__init__(self, parent)
        self.propss = propss
        self.cls = cls

    def run(self):
        for props in self.propss:
            if not props.isfile:
                continue
            f = None
            try:
                if self.cls:
                    f = self.cls()
                    f.load(props.filepath)
                else:
                    f = f311.load_any_file(props.filepath)
                props.f = f
            except Exception as e:
                # Suppresses exceptions
                props.error_message = str(e)
                a99.get_python_logger().exception("Loading file '{}'".format(props.filepath))
                f = None
            finally:
                props.flag_scanned = True
                props.flag_error = f is None
                props.flag_text = a99.is_text_file(props.filepath)

#
# class VisThread(QThread):
#     def __init__(self, parent, vis_class, props):
#         QThread.__init__(self, parent)
#         self.vis_class = vis_class
#         self.props = props
#         self.form = parent
#
#     def run(self):
#         vis_class = self.vis_class
#         props = self.props
#         if vis_class == "txt":
#             w = XText(self.form, open(props.filepath, "r").read(), props.filepath)
#             w.show()
#         else:
#             vis_class().use(props.f)

###############################################################################


class XExplorer(QMainWindow):
    """"F311 Explorer" window

    Args:
      parent=None
      dir_=".": directory name
      _flag_set_dir=True: will not set directory if False (testing purpose only)
    """

    def __init__(self, parent=None, dir_=".", _flag_set_dir=True):
        QMainWindow.__init__(self, parent)
        self.dir = None
        # Whether to close matplotlib plots when window is closed
        self.flag_close_mpl_plots_on_close = True
        self.__vis_classes = []
        # List of (DataFile objects)/(None if not loaded)/(-1 if error loading)
        self.__propss = []
        # # More internals
        self.__flag_loading = False
        self.__flag_visualizing = False
        self.__flag_updating_table = False
        self.__load_thread = None
        self.__flag_creating = True
        self.__lock_propss = a99.MyLock(flag_verbose=False)

        # # Timer to watch for changed files
        t = self.timer_changed = QTimer()
        t.setInterval(1000)  # miliseconds
        t.timeout.connect(self.on_timer_changed_timeout)
        t.start()


        # # Menu bar
        b = self.menuBar()
        m = self.menu_file = b.addMenu("&File")
        ac = self.action_load = m.addAction("Attempt &load")
        ac.setShortcut("Ctrl+L;Ctrl+O")
        ac.triggered.connect(self.on_load)
        ac = self.action_load_as = m.addAction("Load &As...")
        ac.setShortcut("Alt+A")
        ac.triggered.connect(self.on_load_as)
        m.addSeparator()
        ac = m.addAction("&Refresh")
        ac.setShortcut("Ctrl+R")
        ac.triggered.connect(self.on_refresh)
        m.addSeparator()
        ac = m.addAction("Change &directory...")
        ac.setShortcut("Ctrl+D")
        ac.triggered.connect(self.on_cd)
        m.addSeparator()
        ac = m.addAction("&Quit")
        ac.setShortcut("Ctrl+Q")
        ac.triggered.connect(self.close)

        m = self.menu_tools = b.addMenu("&Tools")
        accw = m.addAction("&Collect errors")
        accw.triggered.connect(self.on_collect_errors)

        # # Status bar
        self.labelStatus = QLabel("Welcome")
        sb = self.statusBar()
        sb.insertWidget(0, self.labelStatus, 0)


        # # Central widget

        cw = self.centralWidget = QWidget()
        lcw = self.centralLayout = QVBoxLayout(cw)
        a99.set_margin(lcw, 1)

        # ## Toolbar
        l0 = self.layoutToolbar = QHBoxLayout()


        a0 = self.c88532 = QLabel("<b>&Directory:</b>")
        l0.addWidget(a0)
        #
        a1 = self.lineEditDir = QLineEdit()
        l0.addWidget(a1)
        a1.installEventFilter(self)  # we want _down arrow_ to jump to table widget
        a0.setBuddy(a1)
        a1.returnPressed.connect(self.on_cd2)
        #
        w = self.line44412 = QFrame() # vertical line
        l0.addWidget(w)
        # w.setGeometry(QRect(240, 240, 3, 61))
        w.setFrameShape(QFrame.VLine)
        w.setFrameShadow(QFrame.Sunken)
        w.setFixedWidth(3)
        #
        b6 = self.pushButtonCollectErrors = QPushButton("&Collect errors")
        b6.setToolTip("Searches for errors in log files and reports these errors in a new window.")
        b6.clicked.connect(accw.triggered)
        l0.addWidget(b6)
        #
        lcw.addLayout(l0)



        # # Horizontal splitter
        # Containis a table on the left and a "possibilities are" on the right

        sp = self.splitter = QSplitter(Qt.Horizontal)

        # ## File listing table

        w = self.cdf856 = QWidget()
        l0 = self.c49g7s = QVBoxLayout(w)
        a99.set_margin(l0, 0)
        #
        y = self.c88ss8 = QLabel("<b>Directory &contents</b>")
        l0.addWidget(y)
        #
        t = self.tableWidget = QTableWidget()
        y.setBuddy(t)
        l0.addWidget(t)
        t.installEventFilter(self)
        # t.setSelectionBehavior(QAbstractItemView.SelectRows)
        # t.currentCellChanged.connect(self.on_tableWidget_currentCellChanged)
        t.cellDoubleClicked.connect(self.on_tableWidget_cellDoubleClicked)
        t.setEditTriggers(QTableWidget.NoEditTriggers)
        t.setFont(a99.MONO_FONT)
        t.resizeColumnsToContents()
        t.installEventFilter(self)
        t.selectionModel().selectionChanged.connect(self.selectionChanged)
        t.setContextMenuPolicy(Qt.CustomContextMenu)
        t.customContextMenuRequested.connect(self.on_tableWidget_customContextMenuRequested)

        sp.addWidget(w)

        # ## "Possibilities area"
        # The area is a widget that will be added to the main splitter.
        # The area will contain a vertical layout with two widgets: a label
        # and another splitter.

        w = QWidget()
        sp.addWidget(w)
        l1 = QVBoxLayout(w)
        a99.set_margin(l1, 0)

        # ### Label containing information about the current selection in the table widget

        y = self.c77ww8 = QLabel("<b>Brief</b>")
        l1.addWidget(y)

        x = self.labelSummary = QLabel()
        l1.addWidget(x)

        # ### Splitter containing a "visualization options area"
        #     and a "general file info" area
        s0 = QSplitter(Qt.Vertical)
        l1.addWidget(s0)

        # #### "visualization options" area
        w = self.csslff = QWidget()
        s0.addWidget(w)
        l1 = self.c49378 = QVBoxLayout(w)
        a99.set_margin(l1, 0)
        y = self.c88888 = QLabel("<b>&Actions available for selected file(s)</b>")
        l1.addWidget(y)
        x = self.listWidgetVis = QListWidget(self)
        x.installEventFilter(self)
        x.itemDoubleClicked.connect(self.on_listWidgetVis_itemDoubleClicked)
        y.setBuddy(x)

        l1.addWidget(x)

        # #### "general file info" area
        w = self.c77777 = QWidget()
        s0.addWidget(w)
        l1 = self.c86888 = QVBoxLayout(w)
        a99.set_margin(l1, 0)
        x = self.c83388 = QLabel("<b>Data file &information</b>")
        x.setToolTip("The box below will show information only if the file is of a "
                     "supported type")
        l1.addWidget(x)
        x = self.textEditInfo = QTextEdit(self)
        self.c83388.setBuddy(x)
        l1.addWidget(x)
        x.setReadOnly(True)
        x.setFont(a99.MONO_FONT)

        lcw.addWidget(sp)
        self.setCentralWidget(cw)
        s0.setStretchFactor(0, 3)
        s0.setStretchFactor(1, 10)

        if _flag_set_dir:
            self.set_dir(dir_)

        self.__flag_creating = False

        a99.place_center(self, 1000, 620)

    def set_dir(self, dir_):
        """Sets directory, auto-loads, updates all GUI contents."""

        self.__lock_set_dir(dir_)
        self.__lock_auto_load()
        self.__lock_update_table()
        self.__update_info()
        self.__update_window_title()

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # *
    # Qt override

    def eventFilter(self, obj, event):
        if self.__flag_creating:
            return False

        if obj == self.tableWidget:
            if a99.check_return_space(event, self.on_load):
                return True
            if event.type() == QEvent.KeyPress:
                if event.key() == Qt.Key_Up:
                    if self.tableWidget.currentRow() == 0:
                        self.lineEditDir.setFocus()
        elif obj == self.listWidgetVis:
            return a99.check_return_space(event, self.__visualize)
        elif obj == self.lineEditDir:
            if event.type() == QEvent.KeyPress:
                if event.key() == Qt.Key_Down:
                    self.tableWidget.setFocus()
        return False

    def closeEvent(self, event):
        if self.flag_close_mpl_plots_on_close:
            plt.close("all")

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # *
    # Slots

    def on_tableWidget_cellDoubleClicked(self, row, col):
        # Assuming that double-click will also select the row of the file to load
        self.on_load()

    def on_listWidgetVis_itemDoubleClicked(self, item):
        # Assuming that double-click will also select the row
        try:
            self.__visualize()
        except Exception as e:
            a99.show_error(str(e))
            raise

    def on_tableWidget_customContextMenuRequested(self, position):
        menu = QMenu()
        menu.addAction(self.action_load)
        menu.addAction(self.action_load_as)
        action = menu.exec_(self.tableWidget.mapToGlobal(position))

    def selectionChanged(self, *args):
        if self.__flag_updating_table:
            return
        self.__update_info()

    def on_load(self, _=None):
        if not self.__flag_loading:
            pp = self.__lock_get_current_propss()
            if len(pp) == 1 and not pp[0].isfile:
                self.set_dir(pp[0].filepath)
            else:
                self.__lock_load()

    def on_load_as(self, _=None):
        if not self.__flag_loading:
            pp = self.__lock_get_current_propss()
            if len(pp) == 1 and not pp[0].isfile:
                return
            else:
                form = f311.XSelectDataFile()
                if form.exec_():
                    self.__lock_load(form.cls)

    def on_refresh(self, _=None):
        if not self.__flag_loading:
            self.set_dir(self.dir)

    def on_cd(self, _=None):
        if not self.__flag_loading:
            dir_ = QFileDialog.getExistingDirectory(None, "Change directory", self.dir)[0]
            if dir_:
                self.set_dir(str(dir_))

    def on_cd2(self):
        new_dir = str(self.lineEditDir.text())
        if os.path.isdir(new_dir):
            new_dir = os.path.abspath(new_dir)
            self.set_dir(new_dir)
        else:
            a99.show_error("Invalid directory")

    def on_collect_errors(self, _=None):
        if self.__flag_loading:
            return

        from f311 import explorer as ex

        self.__flag_loading = True
        self.__set_status_text("Collecting errors, please wait...")
        try:
            k = a99.ErrorCollector()
            k.collect_errors(self.dir)
            w = ex.XHTML(self, k.get_html(), "Errors in '{0!s}' and subdirectories".format(self.dir))
            w.show()
        except Exception as e:
            MSG = "Could not collect errors"
            a99.get_python_logger().exception(MSG)
            a99.show_error("{0!s}: {1!s}".format(MSG, str(e)))
        finally:
            self.__flag_loading = False
            self.__set_status_text("")

    def on_timer_changed_timeout(self):
        if self.__flag_loading:
            return
        self.__lock_check_dir_contents_changed()

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # *
    # Internals

    def __update_window_title(self):
        full_dir = os.path.abspath(self.dir)
        self.setWindowTitle("F311 Explorer")
        self.lineEditDir.setText(full_dir)

    def __get_current_vis_class(self):
        return self.__vis_classes[self.listWidgetVis.currentRow()]

    def __update_info(self):
        """Updates "visualization options" and "file info" areas."""

        from f311 import explorer as ex
        import f311

        t = self.tableWidget
        z = self.listWidgetVis
        z.clear()
        classes = self.__vis_classes = []
        propss = self.__lock_get_current_propss()
        npp = len(propss)
        s0, s1 = "", ""
        if npp == 1:
            p = propss[0]
            # Visualization options
            if p.flag_scanned:
                if isinstance(p.f, f311.DataFile):
                    classes.extend(f311.get_suitable_vis_classes(p.f))
                    if ex.VisPrint in classes:
                        classes.remove(ex.VisPrint)
                if p.flag_text:
                    # This is an exception, since "txt" is not a Vis descendant.
                    # This will be properly handled in __visualize()
                    classes.append("txt")

                for x in classes:
                    if x == "txt":
                        text = "View plain text"
                    else:
                        text = x.action
                        text += " ("+x.__name__+")"
                    item = QListWidgetItem(text)
                    z.addItem(item)

            # File info
            s0 = p.get_summary()
            s1 = p.get_info()

        elif npp >= 2:
            s0 = "{0:d} selected".format(npp)
            ff = [p.f for p in propss]
            flag_spectra = all([isinstance(f, f311.FileSpectrum) for f in ff])

            # gambiarra to visualize several PFANT .mod files
            has_pyfant = False
            try:
                import pyfant
                has_pyfant = True
            except:
                pass

            flag_mod = False
            if has_pyfant:
                flag_mod = all([isinstance(f, pyfant.FileModBin) and len(f.records) > 1 for f in ff])

            if flag_spectra:
                z.addItem(QListWidgetItem("Plot spectra stacked"))
                classes.append("sta")
                z.addItem(QListWidgetItem("Plot spectra overlapped"))
                classes.append("ovl")
            elif flag_mod:

                # TODO plugin-based way to handle visualization of multiple selection

                z.addItem(QListWidgetItem("View model grid"))
                classes.append("modgrid")
                
        # File info
        self.labelSummary.setText(s0)
        self.textEditInfo.setPlainText(s1)



    def __set_status_text(self, text):
        self.labelStatus.setText(text)
        QApplication.instance().processEvents()

    def __visualize(self):
        from f311 import explorer as ex
        self.__flag_visualizing = True
        self.__set_status_text("Creating visualization, please wait...")
        try:
            vis_class = self.__get_current_vis_class()
            if vis_class == "ovl":
                propss = self.__lock_get_current_propss()
                spectra = [p.f.spectrum for p in propss]
                ex.plot_spectra_overlapped(spectra)
            elif vis_class == "sta":
                propss = self.__lock_get_current_propss()
                spectra = [p.f.spectrum for p in propss]
                ex.plot_spectra_stacked(spectra)
            elif vis_class == "modgrid":
                propss = self.__lock_get_current_propss()
                models = [p.f for p in propss]
                ex.plot_mod_grid(models)
            else:
                props = self.__lock_get_current_props()
                if vis_class == "txt":
                    w = ex.XText(self, open(props.filepath, "r").read(), props.filepath)
                    w.show()
                else:
                    vis_class().use(props.f)
        finally:
            self.__set_status_text("")
            self.__flag_visualizing = False
        # t = VisThread(self, self.__get_current_vis_class(), self.__get_current_props())
        # t.finished.connect(self.__finished_visualizing)
        # t.start()

    def __finished_loading(self):
        self.__set_status_text("")
        self.__flag_loading = False
        t = self.tableWidget
        row = t.currentRow()
        for props in self.__load_thread.propss:
            i, c = props.row_index, props.get_color()
            for j in range(t.columnCount()):
                t.item(i, j).setBackground(c)
        self.__update_info()

    # def __finished_visualizing(self):
    #     self.__set_status_text("")
    #     self.__flag_visualizing = False

    def __get_auto_size(self):
        # Returns the number of bytes that will be automatically analysed.
        return 100.*2**20  # 100 MB


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # *
    # Internals mutually locked: __lock*()

    def __lock_update_table(self):
        self.__flag_updating_table = True
        try:
            with self.__lock_propss:
                n = len(self.__propss)

                t = self.tableWidget
                a99.reset_table_widget(t, n, 3)
                t.setHorizontalHeaderLabels(["filename", "size", "date modified"])

                for i, p in enumerate(self.__propss):
                    p.row_index = i

                    items = []
                    info = os.stat(p.filepath)
                    filename = os.path.basename(p.filepath)
                    # if not p.isfile:
                    #     filename = "["+filename+"]"
                    item = QTableWidgetItem(filename)
                    items.append(item)
                    t.setItem(i, 0, item)
                    item = QTableWidgetItem(str(info.st_size))
                    items.append(item)
                    item.setTextAlignment(Qt.AlignRight | Qt.AlignVCenter)

                    t.setItem(i, 1, item)
                    item = QTableWidgetItem(time.ctime(info.st_mtime))
                    items.append(item)
                    t.setItem(i, 2, item)

                    color = p.get_color()
#                    if color is None:
#                        color = t.palette().color(t.backgroundRole())
                    if color is not None:
                        for item in items:
                            item.setBackground(color)

            t.resizeColumnsToContents()
        finally:
            self.__flag_updating_table = False

    def __lock_get_current_props(self):
        with self.__lock_propss:
            if len(self.__propss) > 0:
                return self.__propss[self.tableWidget.currentRow()]
            return None

    def __lock_get_current_propss(self):
        with self.__lock_propss:
            ii = self.tableWidget.selectedIndexes()
            i_rows = {idx.row() for idx in ii}
            propss = [self.__propss[i_row] for i_row in i_rows]
            return propss

    def __lock_check_dir_contents_changed(self):
        with self.__lock_propss:
            tw = self.tableWidget
            ir = tw.currentRow()
            ic = tw.currentColumn()

            flag_refresh = False
            filepaths = [p.filepath for p in self.__propss]

            for p in reversed(self.__propss):
                try:
                    t = os.path.getmtime(p.filepath)

                    if t != p.mtime:
                        # File has been changed, mark it as non scanned
                        p.flag_scanned = False
                        p.mtime = t
                        flag_refresh = True
                except OSError as E:
                    if E.errno == 2:
                        # File no longer exists
                        if p.row_index <= ir:
                            ir -= 1
                        self.__propss.remove(p)
                        flag_refresh = True
                    else:
                        raise

            # Newly created files will be added at the bottom and without any sorting
            for filepath in glob.glob(os.path.join(self.dir, "*")):
                if os.path.isfile(filepath):
                    if not filepath in filepaths:
                        p = _FileProps()
                        p.isfile = os.path.isfile(filepath)
                        p.filepath = filepath
                        p.mtime = os.path.getmtime(filepath)
                        self.__propss.append(p)
                        flag_refresh = True

        if flag_refresh:
            self.__lock_update_table()
            self.__update_info()
            if ir < tw.rowCount():
                # Tries to re-position cursor at cell
                tw.setCurrentCell(ir, ic)

    def __lock_set_dir(self, dir_):
        with self.__lock_propss:
            self.dir = os.path.relpath(dir_)
            self.__propss = []

            all_ = glob.glob(os.path.join(self.dir, "*"))
            dirs = list(filter(os.path.isdir, all_))
            dirs.sort()
            files = list(filter(os.path.isfile, all_))
            files.sort()
            dir_ = [os.path.join(self.dir, "..")] + dirs + files

            for i, f in enumerate(dir_):
                filepath = f  # os.path.join(self.dir, f)
                p = _FileProps()
                p.isfile = os.path.isfile(filepath)
                p.filepath = filepath
                p.mtime = os.path.getmtime(filepath)
                self.__propss.append(p)

    def __lock_auto_load(self):
        dir_ = glob.glob(os.path.join(self.dir, "*"))
#        dir_ = [x for x in dir_ if os.path.isfile(x)]
        sizes = [os.path.getsize(x) for x in dir_ if os.path.isfile(x)]
        sizes.sort()
        acc = np.cumsum(sizes)
        threshold = self.__get_auto_size()
        ww = np.where(acc <= threshold)
        if len(ww[0]) == 0:
            return
        max_size = min(sizes[ww[0][-1]], MAX_FILESIZE_AUTO_LOAD)
        # np.savetxt("brinca", sizes)
        # total = sum(sizes)
        # percentile = threshold/total
        #max_size = np.percentile(sizes, percentile)*100
        a99.get_python_logger().info("AUTO LOAD: max file size: {0:g}".format(max_size))
        propss = []
        with self.__lock_propss:
            for p in self.__propss:
                if p.isfile and os.path.getsize(p.filepath) <= max_size:
                    propss.append(p)

        # **Note** this was once crashing because Python modules were being imported
        # without filetype check. FilePy was implemented to be a base class for all
        # ".py-based" file types
        if len(propss) > 0:
            self.__set_status_text("Loading all files sized <= {0:g} kB...".format((max_size/1024)))
            self.__flag_loading = True
            t = self.__load_thread = LoadThread(self, propss)
            t.finished.connect(self.__finished_loading, Qt.QueuedConnection)
            t.start()

    def __lock_load(self, cls=None):
        propss = self.__lock_get_current_propss()
        if len(propss) > 0:
            with self.__lock_propss:
                self.__flag_loading = True
                self.__set_status_text("Loading file(s), please wait...")
                t = self.__load_thread = LoadThread(self, propss, cls)
                t.finished.connect(self.__finished_loading)
                t.start()

#



"""
Ancestor class for all classes that represent an input file.
"""

import os
import a99


__all__ = ["DataFile"]


class DataFile(a99.AttrsPart):
    """
    Class representing a file in disk

    **Attention** For subclasses, filetype check is **strongly advised**. Two ways of doing this:
                      (a) inherit _test_magic() (recommended);
                      (b) if there are no testable magic characters, test for absurd
                          within _do_load(). Try to crash early.
    """
    # Descendants shoulds set this
    default_filename = None
    # Whether it is a text file format (otherwise binary)
    flag_txt = True
    # Whether or not to be considered by load_any_file()
    flag_collect = True
    # List of script names that can edit this file type
    editors = None


    @a99.classproperty
    def description(cls):
        return a99.get_obj_doc0(cls)

    def __init__(self):
        a99.AttrsPart.__init__(self)
        # File name is set by load()
        self.__flag_loaded = False
        self.filename = None

    # # Methods to be implemented by subclasses
    #   =======================================

    def _do_save_as(self, filename):
        raise NotImplementedError()

    def _do_load(self, filename):
        raise NotImplementedError("Forgot to implement _do_load() for class '{}'".
                                  format(self.classname))

    def _test_magic(self, filename):
        """
        Opens file just to verify whether it is what it is expected to be (**raise if not**)

        Implement this if you want to implement file type verification separate from _do_load()

        """
        pass

    # # Interface
    #   =========

    def save_as(self, filename=None):
        """
        Dumps object contents into file on disk.

        Args:
          filename (optional): defaults to self.filename. If passed, self.filename
            will be updated to filename.
        """
        if filename is None:
            filename = self.filename
        if filename is None:
            filename = self.default_filename
        if filename is None:
            raise RuntimeError("Class '{}' has no default filename".format(self.__class__.__name__))
        self._do_save_as(filename)
        self.filename = filename

    def load(self, filename=None):
        """Loads file and registers filename as attribute."""
        assert not self.__flag_loaded, "File can be loaded only once"
        if filename is None:
            filename = self.default_filename
        assert filename is not None, \
            "{0!s} class has no default filename".format(self.__class__.__name__)

        # Convention: trying to open empty file is an error,
        # because it could be of (almost) any type.

        size = os.path.getsize(filename)
        if size == 0:
            raise RuntimeError("Empty file: '{0!s}'".format(filename))

        self._test_magic(filename)
        self._do_load(filename)
        self.filename = filename
        self.__flag_loaded = True

    def init_default(self):
        """
        Initializes object with its default values

        Tries to load self.default_filename from default
        data directory. For safety, filename is reset to None so that it doesn't point to the
        original file.
        """
        import f311
        if self.default_filename is None:
            raise RuntimeError("Class '{}' has no default filename".format(self.__class__.__name__))
        fullpath = f311.get_default_data_path(self.default_filename, class_=self.__class__)
        self.load(fullpath)
        self.filename = None

    def validate(self):
        pass


from .datafile import DataFile
import a99
import shutil
import os


__all__ = ["FileSQLiteDB"]


class FileSQLiteDB(DataFile):
    """Represents a SQLite database file.

    This class is not supposed to be instantialized. It serves as an ancestor for other classes
    that implement specific database schemas."""

    attrs = ["classname", "filename", "tables"]
    flag_txt = False

    # (xgear info): {(table name): {(field name): [caption, description], ...}, ...}
    #
    #  See ftpyfant.convmol.moldb for example
    gui_info = {}

    @property
    def tables(self):
        """Return a list of the database table names"""
        return self._get_table_names()

    def __init__(self):
        DataFile.__init__(self)
        self._conn = None

    # # Overridable
    #   ===========

    def _get_conn(self, filename):
        return a99.get_conn(filename)

    def _create_schema(self):
        """Responsible for executing the CREATE TABLE statements

        Does not actually raise. If you don't want a schema, that's fine"""
        pass

    def _populate(self):
        """Responsible for that initial kick to a DB with all tables created, but empty

        Does not actually raise. If you don't want to populate the database, that's fine"""
        pass

    # # Override
    #   ========

    def init_default(self):
        """Overriden to take default database and save locally

        The issue was that init_default() sets self.filename to None; however there can be no
        SQLite database without a corresponding file (not using *memory* here)

        Should not keep default file open either (as it is in the API directory and shouldn't be
        messed by user)
        """
        import f311
        if self.default_filename is None:
            raise RuntimeError("Class '{}' has no default filename".format(self.__class__.__name__))
        fullpath = f311.get_default_data_path(self.default_filename, class_=self.__class__)
        self.load(fullpath)
        name, ext = os.path.splitext(self.default_filename)
        new = a99.new_filename(os.path.join("./", name), ext)
        self.save_as(new)

    def _do_load(self, filename):
        """Loading here is limited to opening a connection"""
        self.__get_conn(filename=filename)
        self._get_table_names()  # tests if it is really a database

    def _do_save_as(self, filename):
        """Closes connection, copies DB file, and opens again pointing to new file

        **Note** if filename equals current filename, does nothing!
        """
        if filename != self.filename:
            self._ensure_filename()
            self._close_if_open()
            shutil.copyfile(self.filename, filename)
            self.__get_conn(filename=filename)

    def _close_if_open(self):
        if self._conn_is_open():
            self._conn.close()
            self._conn = None

    def _ensure_filename(self):
        if self.filename is None:
            raise RuntimeError("'filename' attribute is not set")

    # # Interface
    #   =========

    def _get_table_names(self):
        return a99.get_table_names(self.__get_conn())

    def delete(self):
        """Removes .sqlite file. **CAREFUL** needless say"""
        self._ensure_filename()
        self._close_if_open()
        os.remove(self.filename)

    def create_schema(self):
        """Creates database schema"""
        self._ensure_filename()
        self._create_schema()

    def populate(self):
        """Series of INSERT statements to populate database with its initial contents"""
        self._ensure_filename()
        self._populate()

    def get_conn(self, flag_force_new=False):
        """Returns sqlite3.Connection object with a non-default row factory.

        Args:
            flag_force_new: returns a new connection irrespective of one already being open.
                            Does **not** close existing open connection

        """
        self._ensure_filename()
        return self.__get_conn(flag_force_new)

    def get_column_names(self, tablename):
        info = self.get_table_info(tablename)
        return list(info.keys())

    def get_table_info(self, tablename):
        """Returns information about fields of a specific table

        Returns:  OrderedDict(("fieldname", MyDBRow), ...))

        **Note** Fields "caption" and "tooltip" are added to rows using information in moldb.gui_info

        """
        conn = self.__get_conn()
        ret = a99.get_table_info(conn, tablename)

        if len(ret) == 0:
            raise RuntimeError("Cannot get info for table '{}'".format(tablename))

        more = self.gui_info.get(tablename)
        for row in ret.values():
            caption, tooltip = None, None
            if more:
                info = more.get(row["name"])
                if info:
                    caption, tooltip = info
            row["caption"] = caption
            row["tooltip"] = tooltip

        return ret

    # # Internal gear
    #   =============

    def _conn_is_open(self):
        """Tests sqlite3 connection, returns T/F"""
        return a99.conn_is_open(self._conn)

    def __get_conn(self, flag_force_new=False, filename=None):
        """Returns connection to database. Tries to return existing connection, unless flag_force_new

        Args:
            flag_force_new:
            filename:

        Returns: sqlite3.Connection object

        **Note** this is a private method because you can get a connection to any file, so it has to
                 be used in the right moment
        """
        flag_open_new = flag_force_new or not self._conn_is_open()

        if flag_open_new:
            if filename is None:
                filename = self.filename
            # funny that __get_conn() calls _get_conn() but that's it
            conn = self._get_conn(filename)
            self._conn = conn
        else:
            conn = self._conn
        return conn




#!/usr/bin/env python

"""
Lists all programs available
"""

import argparse
import os
from collections import OrderedDict
import glob
import sys
import a99
import f311
import re

# These values must match those of the same variables in <project-root>/docs/gen-script-pages.py
SUBDIR = "autoscripts"
PREFIX_EDITABLE = "script-"


def _add_PFANT(d):
    # This can be called (without much shame) a "gambiarra"
    import pyfant as pf
    d["PFANT"] = {"description": "Spectral synthesis-related Fortran binaries",
                  "exeinfo": pf.get_fortrans()}

def _get_programs_dict(pkgname_only, flag_protected, flag_no_pfant=False):
    """Returns dictionary {(package description): [ExeInfo0, ...], ...}"""

    allinfo = f311.get_programs_dict(pkgname_only, flag_protected)
    if not flag_no_pfant and "pyfant" in allinfo:
        _add_PFANT(allinfo)
    return allinfo


def _format_programs_dict(allinfo, format):
    ret = []
    # will indent listing only if there is more than 1 package listed
    ind = 1 if len(allinfo) > 1 else 0
    for name, dexeinfo in allinfo.items():
        descr = dexeinfo["description"]
        exeinfo = dexeinfo["exeinfo"]
        if len(exeinfo) == 0:
            continue

        title = "{} -- {}".format(name, descr)

        # **Note** Using h3/h4 because ot ReST context of use, but it is possible to change this
        ret.extend(a99.format_h3(title, format) + [""])
        linesp, module_len = a99.format_exe_info(exeinfo, format, ind)
        ret.extend(linesp)

    return ret


def _list_packages():
    print("\n".join(a99.format_h1("List of packages available")))
    print("\n".join(f311.COLLABORATORS_S))


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__,  formatter_class=a99.SmartFormatter)
    parser.add_argument('format', type=str, help='Print format', nargs="?", default="text",
                        choices=["text", "markdown-list", "markdown-table", "rest-list", "rest-toctree"])
    parser.add_argument('-p', '--pkgname', type=str, help='List programs from this package only', default="(all)")
    parser.add_argument('-l', '--list-packages', action="store_true", help='Lists all packages')
    parser.add_argument('-k', '--rest-links', action="store_true",
     help='If format=="rest-list", '
          'renders program names as links to their respective documentation pages')
    parser.add_argument('-n', '--no-pfant', action="store_true",
     help='Does not list PFANT binaries')
    parser.add_argument('-_', '--protected', action="store_true", help="Includes protected scripts (starting with '_')")
    args = parser.parse_args()

    pkgname_only = args.pkgname if args.pkgname is not "(all)" else None

    if args.list_packages:
        _list_packages()
        sys.exit()

    if pkgname_only is not None:
        if pkgname_only not in f311.COLLABORATORS_S:
            print("Invalid package name: '{}'\n\n".format(pkgname_only))
            _list_packages()
            sys.exit()

    allinfo = _get_programs_dict(pkgname_only, args.protected, args.no_pfant)
    strlist = _format_programs_dict(allinfo, args.format)


    _out = "\n".join(strlist)

    if args.format.startswith("rest") and args.rest_links:
        out = re.sub("\* ``(.*)\.py``",
                     lambda match: "* :doc:`{0}.py <{1}/{2}{0}>`".format(match.group(1), SUBDIR, PREFIX_EDITABLE),
                     _out)
    else:
        out = _out

    print(out)



from collections import OrderedDict
import a99
import f311
from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

__all__ = ["XSelectDataFile"]

################################################################################
class XSelectDataFile(QDialog):
    """Window to select a DataFile descendant"""

    @property
    def cls(self):
        return self._cls

    def __init__(self, parent=None):
        QDialog.__init__(self, parent)

        # Class selected
        self._cls = None

        self._add_stuff()

        self._data = f311.get_filetypes_info(editor_quote="'", flag_leaf=True)
        # builds search strings for filter
        for row in self._data:
            row["search"] = "/".join((str(x).upper() for x in row.values()))

        self._populate()

        a99.nerdify(self)
        a99.place_center(self, 800, 600)
        self.setWindowTitle("Select class")


    def _add_stuff(self):
        # # Main layout
        lmain = self.layout_main = a99.keep_ref(QVBoxLayout(self))

        # ## Toolbar: checkboxes with executables
        l1 = self.layout_filter = QHBoxLayout()
        lmain.addLayout(l1)

        w = self.label_filter = QLabel("<b>Fil&ter:</b>")
        l1.addWidget(w)
        w = self.lineEdit_filter = QLineEdit()
        self.label_filter.setBuddy(w)
        w.textEdited.connect(self._on_filter)
        l1.addWidget(w)
        # w.setFixedWidth(100)
        l1.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))


        a = self.tableWidget = QTableWidget()
        lmain.addWidget(a)
        a.setSelectionMode(QAbstractItemView.SingleSelection)
        a.setSelectionBehavior(QAbstractItemView.SelectRows)
        a.setEditTriggers(QTableWidget.NoEditTriggers)
        a.setAlternatingRowColors(True)
        a.currentCellChanged.connect(self._on_tableWidget_currentCellChanged)
        a.itemDoubleClicked.connect(self._close_on_double_click)

        # button box with OK and Cancel buttons
        bb = a99.keep_ref(QDialogButtonBox())
        bb.setOrientation(Qt.Horizontal)
        bb.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)
        lmain.addWidget(bb)
        bb.rejected.connect(self.reject)
        bb.accepted.connect(self.accept)


    def _on_tableWidget_currentCellChanged(self):
        idx = self.tableWidget.currentRow()
        self._cls = self._data[idx]["class"]

    def _on_filter(self):
        text = self.lineEdit_filter.text().upper()
        flag_filter = len(text) > 0
        for i, row in enumerate(self._data):
            self.tableWidget.setRowHidden(i, flag_filter and text not in row["search"])

    def _close_on_double_click(self):
        self.accept()

    def _populate(self):
        map = OrderedDict([("classname", "Class name"),
                           ("description", "Description"),
                           ("txtbin", "Text/binary"),
                           ("default_filename", "Default filename"),
                           ("editors", "Editors")])

        nr, nc = len(self._data), len(map)
        t = self.tableWidget
        a99.reset_table_widget(t, nr, nc)
        t.setHorizontalHeaderLabels(map.values())

        for i, row in enumerate(self._data):
            for j, key in enumerate(map):
                item = QTableWidgetItem(row[key])

                t.setItem(i, j, item)


        t.resizeColumnsToContents()



"""
Miscellanea routines that depend on other modules and sub-packages.
"""

import logging
import a99
from astropy.io import fits
from collections import OrderedDict
import textwrap


# TODO usage example for list_data_types(), or even a script

__all__ = [
    "load_any_file", "load_spectrum", "load_spectrum_fits_messed_x",
    "load_with_classes", "load", "get_filetypes_info",
    "filetypes_info_to_rows_header", "tabulate_filetypes_rest"]


def load(filename, class_):
    """Attempts to load file using a specific DataFile descendant.

    Arguments:
      filename -- full path to file
      class_ -- a DataFile descendant, or at least a class having method load() (duck typing)

    Returns: DataFile object if loaded successfully, or None if not.
    """
    return load_with_classes(filename, [class_])


def load_with_classes(filename, classes):
    """Attempts to load file by trial-and-error using a given list of classes.

    Arguments:
      filename -- full path to file
      classes -- list of classes having a load() method

    Returns: DataFile object if loaded successfully, or None if not.

    Note: it will stop at the first successful load.

    Attention: this is not good if there is a bug in any of the file readers,
    because *all exceptions will be silenced!*
    """

    ok = False
    for class_ in classes:
        obj = class_()
        try:
            obj.load(filename)
            ok = True
        # # cannot let IOError through because pyfits raises IOError!!
        # except IOError:
        #     raise
        # # also cannot let OSError through because astropy.io.fits raises OSError!!
        # except OSError:
        #     raise
        except FileNotFoundError:
            raise
        except Exception as e:  # (ValueError, NotImplementedError):
            # Note: for debugging, switch the below to True
            if a99.logging_level == logging.DEBUG:
                a99.get_python_logger().exception("Error trying with class \"{0!s}\"".format(
                                              class_.__name__))
            pass
        if ok:
            break
    if ok:
        return obj
    return None


def load_any_file(filename):
    """
    Attempts to load filename by trial-and-error

    Returns:
        file: A DataFile descendant, whose specific class depends on the file format detected, or None
        if the file canonot be loaded
    """
    import f311

    # Splits attempts using ((binary X text) file) criterion
    if a99.is_text_file(filename):
        return load_with_classes(filename, f311.classes_txt())
    else:
        return load_with_classes(filename, f311.classes_bin())

# def load_any_file_ex(filename):
#     """
#     Attempts to load filename by trial-and-error using _classes as list of classes.
#
#     Returns:
#         file, log_dict: A DataFile descendant, whose specific class depends on the file format detected, or None
#               if the file canonot be loaded
#     """
#     import f311
#
#     # Splits attempts using ((binary X text) file) criterion
#     if a99.is_text_file(filename):
#         return load_with_classes(filename, f311.classes_txt())
#     else:
#         return load_with_classes(filename, f311.classes_bin())


def load_spectrum(filename):
    """
    Attempts to load spectrum as one of the supported types.

    Returns:
        a Spectrum, or None
    """
    import f311

    f = load_with_classes(filename, f311.classes_sp())
    if f:
        return f.spectrum
    return None


def load_spectrum_fits_messed_x(filename, sp_ref=None):
    """Loads FITS file spectrum that does not have the proper headers. Returns a Spectrum"""
    import f311.filetypes as ft

    # First tries to load as usual
    f = load_with_classes(filename, (ft.FileSpectrumFits,))

    if f is not None:
        ret = f.spectrum
    else:
        hdul = fits.open(filename)

        hdu = hdul[0]
        if not hdu.header.get("CDELT1"):
            hdu.header["CDELT1"] = 1 if sp_ref is None else sp_ref.delta_lambda
        if not hdu.header.get("CRVAL1"):
            hdu.header["CRVAL1"] = 0 if sp_ref is None else sp_ref.x[0]

        ret = ft.Spectrum()
        ret.from_hdu(hdu)
        ret.filename = filename
        original_shape = ret.y.shape  # Shape of data before squeeze
        # Squeezes to make data of shape e.g. (1, 1, 122) into (122,)
        ret.y = ret.y.squeeze()

        if len(ret.y.shape) > 1:
            raise RuntimeError(
                "Data contains more than 1 dimension (shape is {0!s}), "
                "FITS file is not single spectrum".format(original_shape))

    return ret



FILE_TYPE_INFO_ATTRS = OrderedDict(zip(["description", "default_filename", "classname", "editors"],
                                       ["Description", "Default filename", "Class name", "Editors"]))


def get_filetypes_info(editor_quote="`", flag_leaf=True):
    """
    Reports available data types

    Args:
        editor_quote: character to enclose the name of the editor script between.
        flag_leaf: see tabulate_filetypes_rest()

    Returns:
        list: list of FileTypeInfo
    """
    NONE_REPL = ""
    import f311
    data = []  # [FileTypeInfo, ...]

    for attr in f311.classes_file(flag_leaf):
        description = a99.get_obj_doc0(attr)

        def_ = NONE_REPL if attr.default_filename is None else attr.default_filename
        ee = attr.editors
        if ee is None:
            ee = NONE_REPL
        else:
            # Example: "``mained.py``, ``x.py``"
            ee = ", ".join(["{0}{1}{0}".format(editor_quote, x, editor_quote) for x in ee])

        data.append({"description": description, "default_filename": def_, "classname": attr.__name__,
                     "editors": ee, "class": attr, "txtbin": "text" if attr.flag_txt else "binary"})

    data.sort(key=lambda x: x["description"])

    return data


def filetypes_info_to_rows_header(infos, attrnames=None, header=None, flag_wrap_description=False,
                                  description_width=40):
    """
    Converts filetype information to a (multiline_rows, header) tuple that can be more easily be tabulated

    **Attention** uses ReST syntax, using a "|br|" marker for line break. It requires the .rst source
                  file to contain the following bit:

        .. |br| raw:: html

           <br />

    Args:
        infos: list of FileTypeInfo
        attrnames: list of attribute names (keys of FILE_TYPE_INFO_ATTRS).
                   Defaults to all attributes
        header: list of strings containing headers. If not passed, uses default names
        flag_wrap_description: whether to wrap the description text
        description_width: width to wrap the description text (effective only if
                           flag_wrap_description is True)

    Returns:
        tuple: (rows, header): rows is a list of lists
    """

    if attrnames is None:
        attrnames = FILE_TYPE_INFO_ATTRS.keys()
    if header is None:
        header = [FILE_TYPE_INFO_ATTRS[key] for key in attrnames]

    if flag_wrap_description:
        wr = textwrap.TextWrapper(width=description_width, subsequent_indent="|br| ")

    data = []
    for i, info in enumerate(infos):
        row = []
        for j, attrname in enumerate(attrnames):
            if attrname != "description" or not flag_wrap_description:
                row.append(info[attrname])
            else:
                row.append(wr.wrap(info[attrname]))
        data.append(row)

    return data, header



#
#     return tabulate.tabulate(data, header)
# a99.markdown_table(("File type description", "Default filen


def tabulate_filetypes_rest(attrnames=None, header=None, flag_wrap_description=True,
                            description_width=40, flag_leaf=True):
    """
    Generates a reST multirow table

    Args:
        attrnames: list of attribute names (keys of FILE_TYPE_INFO_ATTRS).
                   Defaults to all attributes
        header: list of strings containing headers. If not passed, uses default names
        flag_wrap_description: whether to wrap the description text
        description_width: width to wrap the description text (effective only if
                           flag_wrap_description is True)
        flag_leaf: returns only classes that do not have subclasses
                   ("leaf" nodes as in a class tree graph)
    """

    infos = get_filetypes_info(editor_quote="``", flag_leaf=flag_leaf)
    rows, header = filetypes_info_to_rows_header(infos, attrnames, header, flag_wrap_description,
                                                 description_width)
    ret = a99.rest_table(rows, header)
    return ret



"""
Data representations.
"""

from .spectrum import *
from .datafile import *
from .filesqlitedb import *
from .filepy import *
from .filefits import *
from .filespectrum import *
from .filefits import *
from .filespectrum import *



"""Classes FileSpectrum and descendants"""

__all__ = ["FileSpectrum", "FileSpectrumXY", "FileSpectrumFits"]


from .spectrum import Spectrum
import logging
from .datafile import DataFile
from a99 import overwrite_fits, write_lf
from astropy.io import fits
import os
import numpy as np
from .filefits import test_fits_magic

class FileSpectrum(DataFile):
    """Base class for all files representing a single 1D spectrum"""
    attrs = ['spectrum']
    editors = ["splisted.py"]

    def __init__(self):
        DataFile.__init__(self)
        self.spectrum = Spectrum()

    def load(self, filename=None):
        """Method was overriden to set spectrum.filename as well"""
        DataFile.load(self, filename)
        self.spectrum.filename = filename


class FileSpectrumXY(FileSpectrum):
    """
    "Lambda-flux" Spectrum (2-column text file)

    File may have comment lines; these will be ignored altogether, because the file
    is read with numpy.loadtxt()
    """

    def _do_load(self, filename):
        A = np.loadtxt(filename)
        if len(A.shape) < 2:
            raise RuntimeError("File {0!s} does not contain 2D array".format(filename))
        if A.shape[1] < 2:
            raise RuntimeError("File {0!s} must contain at least two columns".format(filename))
        if A.shape[1] > 2:
            logging.warning("File {0!s} has more than two columns, taking only first and second".format(filename))
        sp = self.spectrum = Spectrum()
        sp.x = A[:, 0]
        sp.y = A[:, 1]
#        logging.debug("Just read XY Spectrum '%s'" % filename)


    def _do_save_as(self, filename):
        with open(filename, "w") as h:
            for x, y in zip(self.spectrum.x, self.spectrum.y):
                write_lf(h, "{0:.10g} {1:.10g}".format(x, y))


class FileSpectrumFits(FileSpectrum):
    """FITS Spectrum"""
    flag_txt = False

    def _do_load(self, filename):
        fits_obj = fits.open(filename)

        if len(fits_obj) > 1:
            raise RuntimeError("File has more than 1 frame, probably not a FITS spectrum")

        # fits_obj.info()
        hdu = fits_obj[0]
        hdu.verify()
        sp = self.spectrum = Spectrum()
        sp.from_hdu(hdu)

    def _do_save_as(self, filename):
        """Saves spectrum back to FITS file."""

        if len(self.spectrum.x) < 2:
            raise RuntimeError("Spectrum must have at least two points")
        if os.path.isfile(filename):
            os.unlink(filename)  # PyFITS does not overwrite file

        hdu = self.spectrum.to_hdu()
        overwrite_fits(hdu, filename)

    def _test_magic(self, filename):
        test_fits_magic(filename)




import sys
import os

if sys.version_info[0] < 3:
    print("Python version detected:\n*****\n{0!s}\n*****\nCannot run, must be using Python 3".format(sys.version))
    sys.exit()

from setuptools import setup, find_packages
from glob import glob


def find_scripts(pkgnames):
    ret = []
    for pkgname in pkgnames:
        for item in os.walk(pkgname):
            _, last = os.path.split(item[0])
            if last == "scripts":
                ret.extend(glob(os.path.join(item[0], '*.py')))
    return ret


PACKAGE_NAME = "f311"

pkgs = find_packages()
scripts = find_scripts([PACKAGE_NAME])

setup(
    name=PACKAGE_NAME,
    packages=find_packages(),
    include_package_data=True,
    version='18.1.6.0',
    license='GNU GPLv3',
    platforms='any',
    description='Astronomy-related API, command-line tools, and windowed applications',
    author='Julio Trevisan',
    author_email='juliotrevisan@gmail.com',
    url='https://github.com/trevisanj/f311',
    keywords= ['astronomy', "fits", "spectroscopy", "spectral synthesis", "photometry",
               "honl-london", "nist", "hitran", "multiplicity", "line strength", "kovacs", "franck-condon"],
    install_requires=["numpy", "scipy", "matplotlib", "astropy", "configobj", "bs4", "lxml",
                      "robobrowser", "requests", "fortranformat", "tabulate", "rows", "pyqt5",
                      "a99>=18.1.6.0", "airvacuumvald"],
    scripts=scripts
)


# TODO later install_requires=['numpy', 'matplotlib', 'pyqt5'],  # matplotlib never gets installed correctly by pip, but anyway...



import a99
from .basic import Vis


__all__ = ["VisPrint"]


class VisPrint(Vis):
    """Prints object to screen."""

    input_classes = (object,)
    action = "Print to console"

    def _do_use(self, obj):
        print(obj)



from PyQt5.QtCore import Qt
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import os
from matplotlib import pyplot as plt
import a99
from f311 import filetypes as ft
import f311


__all__ = ["XFileMainWindow", "NullEditor", "XFileMainWindowBase", "MyPage", "NullEditor"]


class MyPage(object):
    """Object holding data related to one tab in a XFileMainWindowBase object"""

    @property
    def flag_op(self):
        return not isinstance(self.editor, NullEditor)

    def __init__(self, text_tab="", flag_changed=False, text_saveas="Save as...",
                 text_load="Load...", cls_save=None, clss_load=(), wild="*.*", editor=None,
                 flag_autosave=False):
        if editor is None:
            editor = NullEditor()

        self.text_tab = text_tab
        self.text_saveas = text_saveas
        self.text_load = text_load
        # save class
        self.cls_save = cls_save
        # accepted load classes
        self.clss_load = clss_load
        # e.g. '*.fits'
        self.wild = wild
        # editor widgets
        self.editor = editor
        # True if the editor saves the file automatically ("(changed)" will never be shown)
        self.flag_autosave = flag_autosave

        self.flag_changed = flag_changed

    def make_text_load(self):
        t = self.text_load
        return t if t is not None else "Load {}...".format(self.cls_save.description)

    def make_text_saveas(self):
        t = self.text_saveas
        return t if t is not None else "Save {} as...".format(self.cls_save.description)


class XFileMainWindowBase(a99.XLogMainWindow):
    """Application template with file operations in two tabs: (application area) and (log)

    Args:
        parent: passed to QMainWindow constructor
        fobjs: sequence of DataFile objects to be loaded by their respective editors
    """

    def __init__(self, parent=None, fobjs=None):
        def keep_ref(obj):
            self._refs.append(obj)
            return obj

        a99.XLogMainWindow.__init__(self, parent)

        self.save_dir = None
        self.load_dir = None

        # [MyPage, ...]
        self.pages = []

        # # Menu bar
        self.__add_menu()

        # # Central layout and widget
        #
        self.centralWidget = cw = QWidget()
        self.setCentralWidget(cw)

        self.centralLayout = lantanide = QVBoxLayout(cw)
        a99.set_margin(lantanide, 1)

        # ## Tabs
        tt = self.tabWidget = QTabWidget(self)
        lantanide.addWidget(tt)
        tt.setFont(a99.MONO_FONT)

        # ## "Last Log" label /  status label
        x = self.label_last_log = QLabel(self)
        # http://stackoverflow.com/questions/6721149/enable-qlabel-to-shrink-even-if-it-truncates-text
        x.setSizePolicy(QSizePolicy(QSizePolicy.Ignored, QSizePolicy.Fixed))
        lantanide.addWidget(x)

        # Custom add stuff (for subclasses)
        self._add_stuff()

        # Adds tabs for pages
        for page in self.pages:
            self.tabWidget.addTab(page.editor, "")

        # Connects changed signals from all editor to keep track of changes
        for page in self.pages:
            page.editor.changed.connect(self._on_changed)

        a99.nerdify(self)

        self.load_many(fobjs)

        self._update_gui_text_tabs()

    def load_many(self, fobjs=None):
        """Loads as many files as the number of pages

        Args:
            fobjs: [filename or DataFile obj, ...]"""
        if fobjs is not None:
            # tolerance
            if not hasattr(fobjs, "__iter__"):
                fobjs = [fobjs]

            for index, (fobj, page) in enumerate(zip(fobjs, self.pages)):
                if fobj is None:
                    continue
                elif isinstance(fobj, ft.DataFile):
                    self.load(fobj, index)
                elif isinstance(fobj, str):
                    self.load_filename(fobj, index)
                else:
                    raise TypeError("Invalid object of class '{}'".format(fobj.__class__.__name__))

    def _add_stuff(self):
        """Responsible for adding menu options, widgets, and pages, setting window title etc.

        Not necessary to add widgets to the tabWidget
        """
        pass

    def __add_menu(self):
        b = self.menuBar()
        m = self.menu_file = b.addMenu("&File")
        ac = m.addAction(a99.get_icon("document-open"), "&Open...")
        ac.setShortcut("Ctrl+O")
        ac.triggered.connect(self.on_load)
        m.addSeparator()
        ac = m.addAction(a99.get_icon("document-save"), "&Save")
        ac.setShortcut("Ctrl+S")
        ac.triggered.connect(self.on_save)
        ac = m.addAction(a99.get_icon("document-save-as"), "Save &as...")
        ac.setShortcut("Ctrl+Shift+S")
        ac.triggered.connect(self.on_save_as)
        ac = m.addAction("Save a&ll...")
        ac.setShortcut("Alt+Shift+S")
        ac.triggered.connect(self.on_save_all)
        m.addSeparator()
        ac = m.addAction("Load &default")
        ac.setStatusTip("Loads default")
        ac.setShortcut("Ctrl+D")
        ac.triggered.connect(self.on_reset)
        m.addSeparator()
        ac = m.addAction(a99.get_icon("system-shutdown"), "&Quit")
        ac.setShortcut("Ctrl+Q")
        ac.triggered.connect(self.close)

    def _add_log_tab(self):
        """Adds element to pages and new tab"""

        # text_tab = "Log (Alt+&{})".format(len(self.pages)+1)
        text_tab = "Log"

        self.pages.append(MyPage(text_tab=text_tab))

        # ### Log tab
        te = self.textEdit_log = self.keep_ref(QTextEdit())
        te.setReadOnly(True)
        self.tabWidget.addTab(te, text_tab)


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Interface

    def load(self, fobj, index=None):
        """
        Loads given DataFile object. **tolerant with None**

        Args:
            fobj: object of one of accepted classes
            index: tab index to load fobj into. If not passed, loads into current tab
        """
        if index is None:
            index = self._get_tab_index()
        page = self.pages[index]

        if fobj is None:
            return


        if not isinstance(fobj, tuple(page.clss_load)):
            raise RuntimeError('Object to load must be in {0!s} (not a {1!s})'.format(
             [x.__name__ for x in page.clss_load], fobj.__class__.__name__))

        page.editor.load(fobj)
        self._update_gui_text_tabs()

    def load_filename(self, filename, index=None):
        """
        Loads file given filename

        Args:
            filename:
            index: tab index to load file into. If not passed, loads into current tab
        """
        filename = str(filename)  # QString protection
        if index is None:
            index = self._get_tab_index()
        page = self.pages[index]

        # Maybe this is set on purpose before loading attempt to leave new load_dir set (?)
        self.load_dir, _ = os.path.split(filename)

        clss = page.clss_load
        if len(clss) == 1:
            # If there is only one class to handle the file, will load it in a way that eventual
            # load errors will raise
            f = clss[0]()
            f.load(filename)
        else:
            # At the moment, the multi-class alternative will not display particular error information
            # if the file does not load
            f = f311.load_with_classes(filename, page.clss_load)
            if f is None:
                raise RuntimeError("Could not load '{0!s}'".format(filename))

        self.load(f, index)

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Qt override

    def closeEvent(self, event):
        flag_exit, ff = True, []
        for page in self.pages:
            ed = page.editor
            if ed and ed.f and page.flag_changed:
                ff.append(ed.f.description)

        if len(ff) > 0:
            s = "Unsaved changes\n  -" + ("\n  -".join(ff)) + "\n\nAre you sure you want to exit?"
            flag_exit = a99.are_you_sure(True, event, self, "Unsaved changes", s)
        if flag_exit:
            plt.close("all")

    def keyPressEvent(self, evt):
        """This handles Ctrl+PageUp, Ctrl+PageDown, Ctrl+Tab, Ctrl+Shift+Tab"""
        incr = 0
        if evt.modifiers() == Qt.ControlModifier:
            n = self.tabWidget.count()
            if evt.key() in [Qt.Key_PageUp, Qt.Key_Backtab]:
                incr = -1
            elif evt.key() in [Qt.Key_PageDown, Qt.Key_Tab]:
                incr = 1
            if incr != 0:
                new_index = self._get_tab_index() + incr
                if new_index < 0:
                    new_index = n - 1
                elif new_index >= n:
                    new_index = 0
                self.tabWidget.setCurrentIndex(new_index)


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Slots for Qt library signals

    # All file operations

    def on_load(self):
        try:
            if self._tab_is_op():
                self.__generic_load()
        except Exception as e:
            self.add_log_error(a99.str_exc(e), True, e)

    def on_save(self):
        try:
            if self._tab_is_op():
                self.__generic_save()
        except Exception as e:
            self.add_log_error(a99.str_exc(e), True, e)

    def on_save_as(self):
        try:
            if self._tab_is_op():
                self.__generic_save_as()
        except Exception as e:
            self.add_log_error(a99.str_exc(e), True, e)

    def on_save_all(self):
        index_save = self._get_tab_index()
        try:
            for index in range(self.tabWidget.count()):
                self.tabWidget.setCurrentIndex(index)
                try:
                    if not self.__generic_save():
                        # breaks if user cancels a "save as" operation
                        break
                except Exception as e:
                    self.add_log_error(a99.str_exc(e), True, e)
        finally:
            self.tabWidget.setCurrentIndex(index_save)

    def on_reset(self):
        if self._tab_is_op():
            editor = self._get_page().editor
            flag_ok = True
            if editor.f:
                descr = self._get_tab_description()
                r = QMessageBox.question(self, "Load default", "Current setup "
                                                               "for %s will be overwritten with a 'default' setup.\n\n"
                                                               "Confirm?" % descr, QMessageBox.Yes | QMessageBox.No,
                                         QMessageBox.Yes)
                flag_ok = r == QMessageBox.Yes
            if flag_ok:
                self._generic_reset()


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Gear


    def _on_changed(self):
        """Slot for changed events"""
        page = self._get_page()
        if not page.flag_autosave:
            page.flag_changed = True
            self._update_gui_text_tabs()

    def _get_tab_index(self):
        """Returns index of current selected tab."""
        return self.tabWidget.currentIndex()

    def _get_page(self):
        """Returns MyPage corresponding to current tab"""
        return self.pages[self._get_tab_index()]

    def _get_tab_description(self):
        """Returns "description" of current tab (tab text without shortcut info)."""
        text = self._get_page().text_tab
        if "(" in text:
            text = text[:text.index("(") - 1]
        text = text[0].lower() + text[1:]
        return text

    def _tab_is_op(self):
        return self._get_page().flag_op

    def _update_gui_text_tabs(self):
        """Iterates through pages to update tab texts"""
        for index, page in enumerate(self.pages):
            self.tabWidget.setTabText(index, "{} (Alt+&{}){}".format(page.text_tab, index+1, (" (changed)" if page.flag_changed else "")))

    def _generic_reset(self):
        page = self._get_page()
        # index = self._get_tab_index()
        # editor, text, cls = self.editors[index], self.text_loads[index], \
        #                     self.clss[index]
        fobj = page.cls_save()
        fobj.init_default()
        page.editor.load(fobj)
        # perhaps it is save to assume that the status is not 'changed' if there is a filename
        # assigned after init_default
        page.flag_changed = fobj.filename is None
        self._update_gui_text_tabs()

    def __generic_save(self):
        """Returns False if user has cancelled a "save as" operation, otherwise True."""
        page = self._get_page()
        f = page.editor.f
        if not f:
            return True
        if not page.editor.flag_valid:
            a99.show_error("Cannot save, {0!s} has error(s)!".format(f.description))
            return True
        if f.filename:
            f.save_as()

            self.add_log("Saved '{}'".format(f.filename))
            page.flag_changed = False
            self._update_gui_text_tabs()

            if hasattr(page.editor, "update_gui_label_fn"):
                page.editor.update_gui_label_fn()  # duck typing
            return True
        else:
            return self.__generic_save_as()

    def __generic_save_as(self):
        """Returns False if user has cancelled operation, otherwise True."""
        page = self._get_page()
        if not page.editor.f:
            return True
        if page.editor.f.filename:
            d = page.editor.f.filename
        else:
            d = os.path.join(self.save_dir if self.save_dir is not None \
                                 else self.load_dir if self.load_dir is not None \
                else ".", page.editor.f.default_filename)
        new_filename = QFileDialog.getSaveFileName(self, page.make_text_saveas(), d, page.wild)[0]
        if new_filename:
            self.save_dir, _ = os.path.split(str(new_filename))
            page.editor.f.save_as(str(new_filename))
            page.flag_changed = False
            self._update_gui_text_tabs()
            page.editor.update_gui_label_fn()
            return True
        return False

    def __generic_load(self):
        page = self._get_page()
        d = self.load_dir if self.load_dir is not None \
            else self.save_dir if self.save_dir is not None \
            else "."
        new_filename = QFileDialog.getOpenFileName(self, page.make_text_load(), d, page.wild+";;*.*")[0]
        if new_filename:
            self.load_filename(new_filename)


class NullEditor(object):
    """Class to fulfill requirement in XFileMainWindow, easiest way out, no bother"""
    f = None


class XFileMainWindow(XFileMainWindowBase):
    """Application template with file operations in two tabs: (application area) and (log)"""

    def __init__(self, *args, **kwargs):
        XFileMainWindowBase.__init__(self, *args, **kwargs)

        self._add_log_tab()
        # TODO make this flexible if needed
        if self.tabWidget.count() > 0:
            self.tabWidget.setCurrentIndex(0)
